#
# -*- coding: utf-8 -*-
# Copyright 2019 Red Hat Inc.
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)
"""
The iosxr interfaces fact class
It is in this file the configuration is collected from the device
for a given resource, parsed, and the facts tree is populated
based on the configuration.
"""

from __future__ import absolute_import, division, print_function


__metaclass__ = type

import re

from copy import deepcopy

from ansible_collections.ansible.netcommon.plugins.module_utils.network.common import utils

from ansible_collections.cisco.iosxr.plugins.module_utils.network.iosxr.argspec.interfaces.interfaces import (
    InterfacesArgs,
)
from ansible_collections.cisco.iosxr.plugins.module_utils.network.iosxr.utils.utils import (
    get_interface_type,
)


class InterfacesFacts(object):
    """The iosxr interfaces fact class"""

    def __init__(self, module, subspec="config", options="options"):
        self._module = module
        self.argument_spec = InterfacesArgs.argument_spec
        spec = deepcopy(self.argument_spec)
        if subspec:
            if options:
                facts_argument_spec = spec[subspec][options]
            else:
                facts_argument_spec = spec[subspec]
        else:
            facts_argument_spec = spec

        self.generated_spec = utils.generate_dict(facts_argument_spec)

    def get_config(self, connection):
        return connection.get_config(flags="interface")

    def populate_facts(self, connection, ansible_facts, data=None):
        """Populate the facts for interfaces
        :param module: the module instance
        :param connection: the device connection
        :param data: previously collected conf
        :rtype: dictionary
        :returns: facts
        """
        objs = []
        if not data:
            data = self.get_config(connection)

        # operate on a collection of resource x
        config = ("\n" + data).split("\ninterface ")
        for conf in config:
            if conf:
                obj = self.render_config(self.generated_spec, conf)
                if obj:
                    objs.append(obj)

        facts = {}
        if objs:
            facts["interfaces"] = []
            params = utils.validate_config(
                self.argument_spec,
                {"config": objs},
            )
            for cfg in params["config"]:
                facts["interfaces"].append(utils.remove_empties(cfg))

        ansible_facts["ansible_network_resources"].update(facts)
        return ansible_facts

    def render_config(self, spec, conf):
        """
        Render config as dictionary structure and delete keys from spec for null values
        :param spec: The facts tree, generated from the argspec
        :param conf: The configuration
        :rtype: dictionary
        :returns: The generated config
        """

        config = deepcopy(spec)
        match = re.search(r"^(\S+)", conf)
        if match:
            intf = match.group(1)
            if match.group(1).lower() == "preconfigure":
                match = re.search(r"^(\S+) (.*)", conf)
                if match:
                    intf = match.group(2)

            if get_interface_type(intf) == "unknown":
                return {}
            # populate the facts from the configuration
            config["name"] = intf
            config["description"] = utils.parse_conf_arg(conf, "description")
            if utils.parse_conf_arg(conf, "speed"):
                config["speed"] = int(utils.parse_conf_arg(conf, "speed"))
            if utils.parse_conf_arg(conf, "mtu"):
                config["mtu"] = int(utils.parse_conf_arg(conf, "mtu"))
            config["duplex"] = utils.parse_conf_arg(conf, "duplex")
            enabled = utils.parse_conf_cmd_arg(conf, "shutdown", False)
            config["enabled"] = enabled if enabled is not None else True

            return utils.remove_empties(config)



#!/usr/bin/python

# Copyright (c) 2013, Evgenii Terechkov
# Written by Evgenii Terechkov <evg@altlinux.org>
# Based on urpmi module written by Philippe Makowski <philippem@mageia.org>

# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later

from __future__ import annotations

DOCUMENTATION = r"""
module: apt_rpm
short_description: APT-RPM package manager
description:
  - Manages packages with C(apt-rpm). Both low-level (C(rpm)) and high-level (C(apt-get)) package manager binaries required.
extends_documentation_fragment:
  - community.general._attributes
attributes:
  check_mode:
    support: none
  diff_mode:
    support: none
options:
  package:
    description:
      - List of packages to install, upgrade, or remove.
      - Since community.general 8.0.0, may include paths to local C(.rpm) files if O(state=installed) or O(state=present),
        requires C(rpm) Python module.
    aliases: [name, pkg]
    type: list
    elements: str
  state:
    description:
      - Indicates the desired package state.
      - The states V(latest) and V(present_not_latest) have been added in community.general 8.6.0.
      - Please note before community.general 11.0.0, V(present) and V(installed) were equivalent to V(latest). This changed
        in community.general 11.0.0. Now they are equivalent to V(present_not_latest).
    choices:
      - absent
      - present
      - present_not_latest
      - installed
      - removed
      - latest
    default: present
    type: str
  update_cache:
    description:
      - Run the equivalent of C(apt-get update) before the operation. Can be run as part of the package installation or as
        a separate step.
      - Default is not to update the cache.
    type: bool
    default: false
  clean:
    description:
      - Run the equivalent of C(apt-get clean) to clear out the local repository of retrieved package files. It removes everything
        but the lock file from C(/var/cache/apt/archives/) and C(/var/cache/apt/archives/partial/).
      - Can be run as part of the package installation (clean runs before install) or as a separate step.
    type: bool
    default: false
    version_added: 6.5.0
  dist_upgrade:
    description:
      - If true performs an C(apt-get dist-upgrade) to upgrade system.
    type: bool
    default: false
    version_added: 6.5.0
  update_kernel:
    description:
      - If true performs an C(update-kernel) to upgrade kernel packages.
    type: bool
    default: false
    version_added: 6.5.0
requirements:
  - C(rpm) Python package (rpm bindings), optional. Required if O(package) option includes local files.
author:
  - Evgenii Terechkov (@evgkrsk)
"""

EXAMPLES = r"""
- name: Install package foo
  community.general.apt_rpm:
    pkg: foo
    state: present

- name: Install packages foo and bar
  community.general.apt_rpm:
    pkg:
      - foo
      - bar
    state: present

- name: Remove package foo
  community.general.apt_rpm:
    pkg: foo
    state: absent

- name: Remove packages foo and bar
  community.general.apt_rpm:
    pkg: foo,bar
    state: absent

# bar will be the updated if a newer version exists
- name: Update the package database and install bar
  community.general.apt_rpm:
    name: bar
    state: present
    update_cache: true

- name: Run the equivalent of "apt-get clean" as a separate step
  community.general.apt_rpm:
    clean: true

- name: Perform cache update and complete system upgrade (includes kernel)
  community.general.apt_rpm:
    update_cache: true
    dist_upgrade: true
    update_kernel: true
"""

import os
import re
import traceback

from ansible.module_utils.basic import (
    AnsibleModule,
    missing_required_lib,
)
from ansible.module_utils.common.text.converters import to_native

RPM_PYTHON_IMPORT_ERROR: str | None
try:
    import rpm
except ImportError:
    HAS_RPM_PYTHON = False
    RPM_PYTHON_IMPORT_ERROR = traceback.format_exc()
else:
    HAS_RPM_PYTHON = True
    RPM_PYTHON_IMPORT_ERROR = None

APT_CACHE = "/usr/bin/apt-cache"
APT_PATH = "/usr/bin/apt-get"
RPM_PATH = "/usr/bin/rpm"
APT_GET_ZERO = "\n0 upgraded, 0 newly installed"
UPDATE_KERNEL_ZERO = "\nTry to install new kernel "
UPDATE_KERNEL_NO_NEW = "There are no available kernels"


def local_rpm_package_name(path):
    """return package name of a local rpm passed in.
    Inspired by ansible.builtin.yum"""

    header = get_local_rpm_header(path)
    if header is None:
        return None

    return to_native(header[rpm.RPMTAG_NAME])


def get_local_rpm_header(path):
    """return rpm header of a local rpm file."""
    ts = rpm.TransactionSet()
    ts.setVSFlags(rpm._RPMVSF_NOSIGNATURES)
    fd = os.open(path, os.O_RDONLY)
    try:
        return ts.hdrFromFdno(fd)
    except rpm.error:
        return None
    finally:
        os.close(fd)


def get_installed_rpm_header(name):
    """return rpm header of an installed package."""
    ts = rpm.TransactionSet()
    return next(ts.dbMatch(rpm.RPMTAG_NAME, name), None)


def query_package(module, name):
    # rpm -q returns 0 if the package is installed,
    # 1 if it is not installed
    rc, out, err = module.run_command([RPM_PATH, "-q", name])
    return rc == 0


def check_package_version(module, name, local_rpm_path):
    # compare installed and candidate version
    # if newest version already installed return True
    # otherwise return False

    if local_rpm_path is not None:
        local_hdr = get_local_rpm_header(local_rpm_path)
        if local_hdr is None:
            module.fail_json(msg=f"Failed to read version from local RPM file: {local_rpm_path}")
        inst_hdr = get_installed_rpm_header(name)
        if inst_hdr is None:
            return False
        result = rpm.versionCompare(inst_hdr, local_hdr)
        return result >= 0
    else:
        rc, out, err = module.run_command([APT_CACHE, "policy", name], environ_update={"LANGUAGE": "C", "LC_ALL": "C"})
        installed = re.split("\n |: ", out)[2]
        candidate = re.split("\n |: ", out)[4]
        return installed >= candidate


def query_package_provides(module, name, allow_upgrade=False):
    # rpm -q returns 0 if the package is installed,
    # 1 if it is not installed
    local_rpm_path = None
    if name.endswith(".rpm"):
        # Likely a local RPM file
        if not HAS_RPM_PYTHON:
            module.fail_json(
                msg=missing_required_lib("rpm"),
                exception=RPM_PYTHON_IMPORT_ERROR,
            )

        local_rpm_path = name
        name = local_rpm_package_name(name)

    rc, out, err = module.run_command([RPM_PATH, "-q", "--provides", name])
    if rc == 0:
        if not allow_upgrade:
            return True
        if check_package_version(module, name, local_rpm_path):
            return True
    return False


def update_package_db(module):
    rc, update_out, err = module.run_command(
        [APT_PATH, "update"], check_rc=True, environ_update={"LANGUAGE": "C", "LC_ALL": "C"}
    )
    return (False, update_out)


def dir_size(module, path):
    total_size = 0
    for cur_path, _dirs, files in os.walk(path):
        for f in files:
            total_size += os.path.getsize(os.path.join(cur_path, f))
    return total_size


def clean(module):
    t = dir_size(module, "/var/cache/apt/archives")
    rc, out, err = module.run_command([APT_PATH, "clean"], check_rc=True)
    return (t != dir_size(module, "/var/cache/apt/archives"), out)


def dist_upgrade(module):
    rc, out, err = module.run_command(
        [APT_PATH, "-y", "dist-upgrade"], check_rc=True, environ_update={"LANGUAGE": "C", "LC_ALL": "C"}
    )
    return (APT_GET_ZERO not in out, out)


def update_kernel(module):
    rc, out, err = module.run_command(
        ["/usr/sbin/update-kernel", "-y"], environ_update={"LANGUAGE": "C", "LC_ALL": "C"}
    )
    if rc != 0:
        if UPDATE_KERNEL_NO_NEW in err:
            return (False, out)
        module.fail_json(msg=f"Error while updating kernel: {err or out}", rc=rc, stdout=out, stderr=err)
    return (UPDATE_KERNEL_ZERO not in out, out)


def remove_packages(module, packages):
    if packages is None:
        return (False, "Empty package list")

    remove_c = 0
    # Using a for loop in case of error, we can report the package that failed
    for package in packages:
        # Query the package first, to see if we even need to remove
        if not query_package(module, package):
            continue

        rc, out, err = module.run_command(
            [APT_PATH, "-y", "remove", package], environ_update={"LANGUAGE": "C", "LC_ALL": "C"}
        )

        if rc != 0:
            module.fail_json(msg=f"failed to remove {package}: {err}")

        remove_c += 1

    if remove_c > 0:
        return (True, f"removed {remove_c} package(s)")

    return (False, "package(s) already absent")


def install_packages(module, pkgspec, allow_upgrade=False):
    if pkgspec is None:
        return (False, "Empty package list")

    packages = []
    for package in pkgspec:
        if not query_package_provides(module, package, allow_upgrade=allow_upgrade):
            packages.append(package)

    if packages:
        command = [APT_PATH, "-y", "install"] + packages
        rc, out, err = module.run_command(command, environ_update={"LANGUAGE": "C", "LC_ALL": "C"})

        installed = True
        for package in pkgspec:
            if not query_package_provides(module, package, allow_upgrade=False):
                installed = False

        # apt-rpm always have 0 for exit code if --force is used
        if rc or not installed:
            module.fail_json(msg=f"'{' '.join(command)}' failed: {err}")
        else:
            return (True, f"{packages} present(s)")
    else:
        return (False, "Nothing to install")


def main():
    module = AnsibleModule(
        argument_spec=dict(
            state=dict(
                type="str",
                default="present",
                choices=["absent", "installed", "present", "removed", "present_not_latest", "latest"],
            ),
            update_cache=dict(type="bool", default=False),
            clean=dict(type="bool", default=False),
            dist_upgrade=dict(type="bool", default=False),
            update_kernel=dict(type="bool", default=False),
            package=dict(type="list", elements="str", aliases=["name", "pkg"]),
        ),
    )

    if not os.path.exists(APT_PATH) or not os.path.exists(RPM_PATH):
        module.fail_json(msg="cannot find /usr/bin/apt-get and/or /usr/bin/rpm")

    p = module.params

    modified = False
    output = ""

    if p["update_cache"]:
        update_package_db(module)

    if p["clean"]:
        (m, out) = clean(module)
        modified = modified or m

    if p["dist_upgrade"]:
        (m, out) = dist_upgrade(module)
        modified = modified or m
        output += out

    if p["update_kernel"]:
        (m, out) = update_kernel(module)
        modified = modified or m
        output += out

    packages = p["package"]
    if p["state"] in ["installed", "present", "present_not_latest", "latest"]:
        (m, out) = install_packages(module, packages, allow_upgrade=p["state"] == "latest")
        modified = modified or m
        output += out

    if p["state"] in ["absent", "removed"]:
        (m, out) = remove_packages(module, packages)
        modified = modified or m
        output += out

    # Return total modification status and output of all commands
    module.exit_json(changed=modified, msg=output)


if __name__ == "__main__":
    main()



#!/usr/bin/python

# Copyright (c) 2017, Thomas Caravia <taca@kadisius.eu>
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later

from __future__ import annotations

DOCUMENTATION = r"""
module: nosh
author:
  - "Thomas Caravia (@tacatac)"
short_description: Manage services with nosh
description:
  - Control running and enabled state for system-wide or user services.
  - BSD and Linux systems are supported.
extends_documentation_fragment:
  - community.general._attributes
attributes:
  check_mode:
    support: full
  diff_mode:
    support: none
options:
  name:
    type: str
    required: true
    description:
      - Name of the service to manage.
  state:
    type: str
    choices: [started, stopped, reset, restarted, reloaded]
    description:
      - V(started)/V(stopped) are idempotent actions that do not run commands unless necessary.
      - V(restarted) always bounces the service.
      - V(reloaded) sends a SIGHUP or starts the service.
      - V(reset) starts or stops the service according to whether it is enabled or not.
  enabled:
    type: bool
    description:
      - Enable or disable the service, independently of C(*.preset) file preference or running state. Mutually exclusive with
        O(preset). It takes effect prior to O(state=reset).
  preset:
    type: bool
    description:
      - Enable or disable the service according to local preferences in C(*.preset) files. Mutually exclusive with O(enabled).
        Only has an effect if set to true. It takes effect prior to O(state=reset).
  user:
    default: false
    type: bool
    description:
      - Run system-control talking to the calling user's service manager, rather than the system-wide service manager.
requirements:
  - A system with an active nosh service manager, see Notes for further information.
notes:
  - Information on the nosh utilities suite may be found at U(https://jdebp.eu/Softwares/nosh/).
"""

EXAMPLES = r"""
- name: Start dnscache if not running
  community.general.nosh:
    name: dnscache
    state: started

- name: Stop mpd, if running
  community.general.nosh:
    name: mpd
    state: stopped

- name: Restart unbound or start it if not already running
  community.general.nosh:
    name: unbound
    state: restarted

- name: Reload fail2ban or start it if not already running
  community.general.nosh:
    name: fail2ban
    state: reloaded

- name: Disable nsd
  community.general.nosh:
    name: nsd
    enabled: false

- name: For package installers, set nginx running state according to local enable settings, preset and reset
  community.general.nosh:
    name: nginx
    preset: true
    state: reset

- name: Reboot the host if nosh is the system manager, would need a "wait_for*" task at least, not recommended as-is
  community.general.nosh:
    name: reboot
    state: started

- name: Using conditionals with the module facts
  tasks:
    - name: Obtain information on tinydns service
      community.general.nosh:
        name: tinydns
      register: result

    - name: Fail if service not loaded
      ansible.builtin.fail:
        msg: "The {{ result.name }} service is not loaded"
      when: not result.status

    - name: Fail if service is running
      ansible.builtin.fail:
        msg: "The {{ result.name }} service is running"
      when: result.status and result.status['DaemontoolsEncoreState'] == "running"
"""

RETURN = r"""
name:
  description: Name used to find the service.
  returned: success
  type: str
  sample: "sshd"
service_path:
  description: Resolved path for the service.
  returned: success
  type: str
  sample: "/var/sv/sshd"
enabled:
  description: Whether the service is enabled at system bootstrap.
  returned: success
  type: bool
  sample: true
preset:
  description: Whether the enabled status reflects the one set in the relevant C(*.preset) file.
  returned: success
  type: bool
  sample: 'False'
state:
  description: Service process run state, V(none) if the service is not loaded and will not be started.
  returned: if state option is used
  type: str
  sample: "reloaded"
status:
  description: A dictionary with the key=value pairs returned by C(system-control show-json) or V(none) if the service is
    not loaded.
  returned: success
  type: dict
  contains:
    After:
      description: []        # FIXME
      returned: success
      type: list
      sample: ["/etc/service-bundles/targets/basic", "../sshdgenkeys", "log"]
    Before:
      description: []        # FIXME
      returned: success
      type: list
      sample: ["/etc/service-bundles/targets/shutdown"]
    Conflicts:
      description: []        # FIXME
      returned: success
      type: list
      sample: []
    DaemontoolsEncoreState:
      description: []        # FIXME
      returned: success
      type: str
      sample: "running"
    DaemontoolsState:
      description: []        # FIXME
      returned: success
      type: str
      sample: "up"
    Enabled:
      description: []        # FIXME
      returned: success
      type: bool
      sample: true
    LogService:
      description: []        # FIXME
      returned: success
      type: str
      sample: "../cyclog@sshd"
    MainPID:
      description: []        # FIXME
      returned: success
      type: int
      sample: 661
    Paused:
      description: []        # FIXME
      returned: success
      type: bool
      sample: 'False'
    ReadyAfterRun:
      description: []        # FIXME
      returned: success
      type: bool
      sample: 'False'
    RemainAfterExit:
      description: []        # FIXME
      returned: success
      type: bool
      sample: 'False'
    Required-By:
      description: []        # FIXME
      returned: success
      type: list
      sample: []
    RestartExitStatusCode:
      description: []        # FIXME
      returned: success
      type: int
      sample: '0'
    RestartExitStatusNumber:
      description: []        # FIXME
      returned: success
      type: int
      sample: '0'
    RestartTimestamp:
      description: []        # FIXME
      returned: success
      type: int
      sample: 4611686019935648081
    RestartUTCTimestamp:
      description: []        # FIXME
      returned: success
      type: int
      sample: 1508260140
    RunExitStatusCode:
      description: []        # FIXME
      returned: success
      type: int
      sample: '0'
    RunExitStatusNumber:
      description: []        # FIXME
      returned: success
      type: int
      sample: '0'
    RunTimestamp:
      description: []        # FIXME
      returned: success
      type: int
      sample: 4611686019935648081
    RunUTCTimestamp:
      description: []        # FIXME
      returned: success
      type: int
      sample: 1508260140
    StartExitStatusCode:
      description: []        # FIXME
      returned: success
      type: int
      sample: 1
    StartExitStatusNumber:
      description: []        # FIXME
      returned: success
      type: int
      sample: '0'
    StartTimestamp:
      description: []        # FIXME
      returned: success
      type: int
      sample: 4611686019935648081
    StartUTCTimestamp:
      description: []        # FIXME
      returned: success
      type: int
      sample: 1508260140
    StopExitStatusCode:
      description: []        # FIXME
      returned: success
      type: int
      sample: '0'
    StopExitStatusNumber:
      description: []        # FIXME
      returned: success
      type: int
      sample: '0'
    StopTimestamp:
      description: []        # FIXME
      returned: success
      type: int
      sample: 4611686019935648081
    StopUTCTimestamp:
      description: []        # FIXME
      returned: success
      type: int
      sample: 1508260140
    Stopped-By:
      description: []        # FIXME
      returned: success
      type: list
      sample: ["/etc/service-bundles/targets/shutdown"]
    Timestamp:
      description: []        # FIXME
      returned: success
      type: int
      sample: 4611686019935648081
    UTCTimestamp:
      description: []        # FIXME
      returned: success
      type: int
      sample: 1508260140
    Want:
      description: []        # FIXME
      returned: success
      type: str
      sample: "nothing"
    Wanted-By:
      description: []        # FIXME
      returned: success
      type: list
      sample: ["/etc/service-bundles/targets/server", "/etc/service-bundles/targets/sockets"]
    Wants:
      description: []        # FIXME
      returned: success
      type: list
      sample: ["/etc/service-bundles/targets/basic", "../sshdgenkeys"]
user:
  description: Whether the user-level service manager is called.
  returned: success
  type: bool
  sample: false
"""


import json

from ansible.module_utils.basic import AnsibleModule
from ansible.module_utils.service import fail_if_missing


def run_sys_ctl(module, args):
    sys_ctl = [module.get_bin_path("system-control", required=True)]
    if module.params["user"]:
        sys_ctl = sys_ctl + ["--user"]
    return module.run_command(sys_ctl + args)


def get_service_path(module, service):
    (rc, out, err) = run_sys_ctl(module, ["find", service])
    # fail if service not found
    if rc != 0:
        fail_if_missing(module, False, service, msg="host")
    else:
        return out.strip()


def service_is_enabled(module, service_path):
    (rc, out, err) = run_sys_ctl(module, ["is-enabled", service_path])
    return rc == 0


def service_is_preset_enabled(module, service_path):
    (rc, out, err) = run_sys_ctl(module, ["preset", "--dry-run", service_path])
    return out.strip().startswith("enable")


def service_is_loaded(module, service_path):
    (rc, out, err) = run_sys_ctl(module, ["is-loaded", service_path])
    return rc == 0


def get_service_status(module, service_path):
    (rc, out, err) = run_sys_ctl(module, ["show-json", service_path])
    # will fail if not service is not loaded
    if err is not None and err:
        module.fail_json(msg=err)
    else:
        json_out = json.loads(out.strip())
        status = json_out[service_path]  # descend past service path header
        return status


def service_is_running(service_status):
    return service_status["DaemontoolsEncoreState"] in {"starting", "started", "running"}


def handle_enabled(module, result, service_path):
    """Enable or disable a service as needed.

    - 'preset' will set the enabled state according to available preset file settings.
    - 'enabled' will set the enabled state explicitly, independently of preset settings.

    These options are set to "mutually exclusive" but the explicit 'enabled' option will
    have priority if the check is bypassed.
    """

    # computed prior in control flow
    preset = result["preset"]
    enabled = result["enabled"]

    # preset, effect only if option set to true (no reverse preset)
    if module.params["preset"]:
        action = "preset"

        # run preset if needed
        if preset != module.params["preset"]:
            result["changed"] = True
            if not module.check_mode:
                (rc, out, err) = run_sys_ctl(module, [action, service_path])
                if rc != 0:
                    module.fail_json(msg=f"Unable to {action} service {service_path}: {out + err}")
            result["preset"] = not preset
            result["enabled"] = not enabled

    # enabled/disabled state
    if module.params["enabled"] is not None:
        if module.params["enabled"]:
            action = "enable"
        else:
            action = "disable"

        # change enable/disable if needed
        if enabled != module.params["enabled"]:
            result["changed"] = True
            if not module.check_mode:
                (rc, out, err) = run_sys_ctl(module, [action, service_path])
                if rc != 0:
                    module.fail_json(msg=f"Unable to {action} service {service_path}: {out + err}")
            result["enabled"] = not enabled
            result["preset"] = not preset


def handle_state(module, result, service_path):
    """Set service running state as needed.

    Takes into account the fact that a service may not be loaded (no supervise directory) in
    which case it is 'stopped' as far as the service manager is concerned. No status information
    can be obtained and the service can only be 'started'.
    """
    # default to desired state, no action
    result["state"] = module.params["state"]
    state = module.params["state"]
    action = None

    # computed prior in control flow, possibly modified by handle_enabled()
    enabled = result["enabled"]

    # service not loaded -> not started by manager, no status information
    if not service_is_loaded(module, service_path):
        if state in ["started", "restarted", "reloaded"]:
            action = "start"
            result["state"] = "started"
        elif state == "reset":
            if enabled:
                action = "start"
                result["state"] = "started"
            else:
                result["state"] = None
        else:
            result["state"] = None

    # service is loaded
    else:
        # get status information
        result["status"] = get_service_status(module, service_path)
        running = service_is_running(result["status"])

        if state == "started":
            if not running:
                action = "start"
        elif state == "stopped":
            if running:
                action = "stop"
        # reset = start/stop according to enabled status
        elif state == "reset":
            if enabled is not running:
                if running:
                    action = "stop"
                    result["state"] = "stopped"
                else:
                    action = "start"
                    result["state"] = "started"
        # start if not running, 'service' module constraint
        elif state == "restarted":
            if not running:
                action = "start"
                result["state"] = "started"
            else:
                action = "condrestart"
        # start if not running, 'service' module constraint
        elif state == "reloaded":
            if not running:
                action = "start"
                result["state"] = "started"
            else:
                action = "hangup"

    # change state as needed
    if action:
        result["changed"] = True
        if not module.check_mode:
            (rc, out, err) = run_sys_ctl(module, [action, service_path])
            if rc != 0:
                module.fail_json(msg=f"Unable to {action} service {service_path}: {err}")


# ===========================================
# Main control flow


def main():
    module = AnsibleModule(
        argument_spec=dict(
            name=dict(type="str", required=True),
            state=dict(type="str", choices=["started", "stopped", "reset", "restarted", "reloaded"]),
            enabled=dict(type="bool"),
            preset=dict(type="bool"),
            user=dict(type="bool", default=False),
        ),
        supports_check_mode=True,
        mutually_exclusive=[["enabled", "preset"]],
    )

    service = module.params["name"]
    result = {
        "name": service,
        "changed": False,
        "status": None,
    }

    # check service can be found (or fail) and get path
    service_path = get_service_path(module, service)

    # get preliminary service facts
    result["service_path"] = service_path
    result["user"] = module.params["user"]
    result["enabled"] = service_is_enabled(module, service_path)
    result["preset"] = result["enabled"] is service_is_preset_enabled(module, service_path)

    # set enabled state, service need not be loaded
    if module.params["enabled"] is not None or module.params["preset"]:
        handle_enabled(module, result, service_path)

    # set service running state
    if module.params["state"] is not None:
        handle_state(module, result, service_path)

    # get final service status if possible
    if service_is_loaded(module, service_path):
        result["status"] = get_service_status(module, service_path)

    module.exit_json(**result)


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-
# Copyright 2020 Red Hat
# GNU General Public License v3.0+
# (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import absolute_import, division, print_function


__metaclass__ = type
DOCUMENTATION = """
    name: validate
    author: Ganesh Nalawade (@ganeshrn)
    version_added: "1.0.0"
    short_description: Validate data with provided criteria
    description:
        - Validate I(data) with provided I(criteria) based on the validation I(engine).
    options:
      data:
        type: raw
        description:
        - A data that will be validated against I(criteria).
        - This option represents the value that is passed to test plugin as check.
          For example C(config_data is ansible.utils.validate(criteria=criteria), in this case B(config_data)
          represents this option.
        - For the type of I(data) that represents this value refer documentation of individual validate plugins.
        required: True
      criteria:
        type: raw
        description:
        - The criteria used for validation of value that represents I(data) options.
        - This option is passed to the test plugin as key, value pair
          For example C(config_data is ansible.utils.validate(criteria=criteria)), in
          this case the value of I(criteria) key represents this criteria for data validation.
        - For the type of I(criteria) that represents this value refer documentation of individual validate plugins.
        required: True
      engine:
        type: str
        description:
        - The name of the validate plugin to use.
        - This option can be passed in test plugin as a key, value pair
          For example C(config_data is ansible.utils.validate(engine='ansible.utils.jsonschema', criteria=criteria)), in
          this case the value of I(engine) key represents the engine to be use for data validation.
          If the value is not provided the default value that is C(ansible.utils.jsonschema) will be used.
        - The value should be in fully qualified collection name format that is
          B(<org-name>.<collection-name>.<validate-plugin-name>).
        default: ansible.utils.jsonschema
    notes:
    - For the type of options I(data) and I(criteria) refer the individual validate plugin
      documentation that is represented in the value of I(engine) option.
    - For additional plugin configuration options refer the individual validate plugin
      documentation that is represented by the value of I(engine) option.
    - The plugin configuration option can be either passed as C(key=value) pairs within test plugin
      or set as environment variables.
    - The precedence the validate plugin configurable option is the variable passed within test plugin
      as C(key=value) pairs followed by task variables followed by environment variables.
"""

EXAMPLES = r"""
- name: set facts for data and criteria
  ansible.builtin.set_fact:
    data: "{{ lookup('ansible.builtin.file', './validate/data/show_interfaces_iosxr.json')}}"
    criteria: "{{ lookup('ansible.builtin.file', './validate/criteria/jsonschema/show_interfaces_iosxr.json')}}"

- name: validate data in json format using jsonschema with test plugin
  ansible.builtin.set_fact:
    is_data_valid: "{{ data is ansible.utils.validate(engine='ansible.utils.jsonschema', criteria=criteria, draft='draft7') }}"
"""

RETURN = """
  _raw:
    description:
      - If data is valid return C(true)
      - If data is invalid return C(false)
"""

from ansible.errors import AnsibleError
from ansible.module_utils.common.text.converters import to_text

from ansible_collections.ansible.utils.plugins.module_utils.common.argspec_validate import (
    check_argspec,
)
from ansible_collections.ansible.utils.plugins.module_utils.common.utils import to_list
from ansible_collections.ansible.utils.plugins.plugin_utils.base.validate import _load_validator


ARGSPEC_CONDITIONALS = {}


def validate(*args, **kwargs):
    if not len(args):
        raise AnsibleError(
            "Missing either 'data' value in test plugin input,"
            "refer ansible.utils.validate test plugin documentation for details",
        )

    params = {"data": args[0]}

    for item in ["engine", "criteria"]:
        if kwargs.get(item):
            params.update({item: kwargs[item]})

    valid, argspec_result, updated_params = check_argspec(
        DOCUMENTATION,
        "validate test",
        schema_conditionals=ARGSPEC_CONDITIONALS,
        **params,
    )
    if not valid:
        raise AnsibleError(
            "{argspec_result} with errors: {argspec_errors}".format(
                argspec_result=argspec_result.get("msg"),
                argspec_errors=argspec_result.get("errors"),
            ),
        )

    validator_engine, validator_result = _load_validator(
        engine=updated_params["engine"],
        data=updated_params["data"],
        criteria=updated_params["criteria"],
        kwargs=kwargs,
    )
    if validator_result.get("failed"):
        raise AnsibleError(
            "validate lookup plugin failed with errors: %s" % validator_result.get("msg"),
        )

    try:
        result = validator_engine.validate()
    except AnsibleError as exc:
        raise AnsibleError(to_text(exc, errors="surrogate_then_replace"))
    except Exception as exc:
        raise AnsibleError(
            "Unhandled exception from validator '{validator}'. Error: {err}".format(
                validator=updated_params["engine"],
                err=to_text(exc, errors="surrogate_then_replace"),
            ),
        )

    errors = to_list(result.get("errors", []))
    if len(errors):
        return False

    return True


class TestModule(object):
    """data validation test"""

    test_map = {"validate": validate}

    def tests(self):
        return self.test_map



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Copyright (c) 2021, Cisco Systems
# GNU General Public License v3.0+ (see LICENSE or
# https://www.gnu.org/licenses/gpl-3.0.txt)

DOCUMENTATION = r"""
module: organizations_sm_sentry_policies_assignments
short_description: Resource module for organizations _sm _sentry _policies _assignments
description:
  - Manage operation update of the resource organizations _sm _sentry _policies _assignments.
    - > Update an Organizations Sentry Policies using the provided list. Sentry Policies
    are ordered in descending order of priority i.e. Highest priority at the bottom,
    this is opposite the Dashboard UI. Policies not present in the request will be
    deleted.
version_added: '1.0.0'
extends_documentation_fragment:
  - cisco.meraki.module
author: Francisco Munoz (@fmunoz)
options:
  items:
    description: Sentry Group Policies for the Organization keyed by Network Id.
    elements: dict
    suboptions:
      networkId:
        description: The Id of the Network.
        type: str
      policies:
        description: Array of Sentry Group Policies for the Network.
        elements: dict
        suboptions:
          groupPolicyId:
            description: The Group Policy Id.
            type: str
          policyId:
            description: The Sentry Policy Id, if updating an existing Sentry Policy.
            type: str
          scope:
            description: The scope of the Sentry Policy.
            type: str
          smNetworkId:
            description: The Id of the Systems Manager Network.
            type: str
          tags:
            description: The tags for the Sentry Policy.
            elements: str
            type: list
        type: list
    type: list
  organizationId:
    description: OrganizationId path parameter. Organization ID.
    type: str
requirements:
  - meraki >= 2.4.9
  - python >= 3.5
seealso:
  - name: Cisco Meraki documentation for sm updateOrganizationSmSentryPoliciesAssignments
    description: Complete reference of the updateOrganizationSmSentryPoliciesAssignments
      API.
    link: https://developer.cisco.com/meraki/api-v1/#!update-organization-sm-sentry-policies-assignments
notes:
  - SDK Method used are
    sm.Sm.update_organization_sm_sentry_policies_assignments,
  - Paths used are
    put /organizations/{organizationId}/sm/sentry/policies/assignments,
"""

EXAMPLES = r"""
- name: Update all
  cisco.meraki.organizations_sm_sentry_policies_assignments:
    meraki_api_key: "{{ meraki_api_key }}"
    meraki_base_url: "{{ meraki_base_url }}"
    meraki_single_request_timeout: "{{ meraki_single_request_timeout }}"
    meraki_certificate_path: "{{ meraki_certificate_path }}"
    meraki_requests_proxy: "{{ meraki_requests_proxy }}"
    meraki_wait_on_rate_limit: "{{ meraki_wait_on_rate_limit }}"
    meraki_nginx_429_retry_wait_time: "{{ meraki_nginx_429_retry_wait_time }}"
    meraki_action_batch_retry_wait_time: "{{ meraki_action_batch_retry_wait_time }}"
    meraki_retry_4xx_error: "{{ meraki_retry_4xx_error }}"
    meraki_retry_4xx_error_wait_time: "{{ meraki_retry_4xx_error_wait_time }}"
    meraki_maximum_retries: "{{ meraki_maximum_retries }}"
    meraki_output_log: "{{ meraki_output_log }}"
    meraki_log_file_prefix: "{{ meraki_log_file_prefix }}"
    meraki_log_path: "{{ meraki_log_path }}"
    meraki_print_console: "{{ meraki_print_console }}"
    meraki_suppress_logging: "{{ meraki_suppress_logging }}"
    meraki_simulate: "{{ meraki_simulate }}"
    meraki_be_geo_id: "{{ meraki_be_geo_id }}"
    meraki_caller: "{{ meraki_caller }}"
    meraki_use_iterator_for_get_pages: "{{ meraki_use_iterator_for_get_pages }}"
    meraki_inherit_logging_config: "{{ meraki_inherit_logging_config }}"
    items:
      - networkId: N_24329156
        policies:
          - groupPolicyId: '1284392014819'
            policyId: '1284392014819'
            scope: withAny
            smNetworkId: N_24329156
            tags:
              - tag1
              - tag2
    organizationId: string
"""
RETURN = r"""
meraki_response:
  description: A dictionary or list with the response returned by the Cisco Meraki Python SDK
  returned: always
  type: dict
  sample: >
    {
      "items": [
        {
          "networkId": "string",
          "policies": [
            {
              "policyId": "string",
              "networkId": "string",
              "smNetworkId": "string",
              "tags": [
                "string"
              ],
              "scope": "string",
              "groupNumber": "string",
              "groupPolicyId": "string",
              "priority": "string",
              "createdAt": "string",
              "lastUpdatedAt": "string"
            }
          ]
        }
      ]
    }
"""



# Copyright (c) 2022 Felix Fontein <felix@fontein.de>
# Copyright 2016 Red Hat | Ansible
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later

# Note that this module util is **PRIVATE** to the collection. It can have breaking changes at any time.
# Do not use this from other collections or standalone plugins/modules!

from __future__ import annotations

import re
import typing as t
from time import sleep

from ansible.module_utils.common.text.converters import to_text

from ansible_collections.community.docker.plugins.module_utils._api.utils.utils import (
    parse_repository_tag,
)
from ansible_collections.community.docker.plugins.module_utils._module_container.base import (
    Client,
)
from ansible_collections.community.docker.plugins.module_utils._util import (
    DifferenceTracker,
    DockerBaseClass,
    compare_generic,
    is_image_name_id,
    normalize_ip_address,
    sanitize_result,
)

if t.TYPE_CHECKING:
    from collections.abc import Sequence

    from ansible.module_utils.basic import AnsibleModule

    from .base import Engine, EngineDriver, Option, OptionGroup


class Container(DockerBaseClass):
    def __init__(
        self, container: dict[str, t.Any] | None, engine_driver: EngineDriver
    ) -> None:
        super().__init__()
        self.raw = container
        self.id: str | None = None
        self.image: str | None = None
        self.image_name: str | None = None
        self.container = container
        self.engine_driver = engine_driver
        if container:
            self.id = engine_driver.get_container_id(container)
            self.image = engine_driver.get_image_from_container(container)
            self.image_name = engine_driver.get_image_name_from_container(container)
        self.log(self.container, pretty_print=True)

    @property
    def exists(self) -> bool:
        return bool(self.container)

    @property
    def removing(self) -> bool:
        return (
            self.engine_driver.is_container_removing(self.container)
            if self.container
            else False
        )

    @property
    def running(self) -> bool:
        return (
            self.engine_driver.is_container_running(self.container)
            if self.container
            else False
        )

    @property
    def paused(self) -> bool:
        return (
            self.engine_driver.is_container_paused(self.container)
            if self.container
            else False
        )


class ContainerManager(DockerBaseClass, t.Generic[Client]):
    def __init__(
        self,
        module: AnsibleModule,
        engine_driver: EngineDriver,
        client: Client,
        active_options: list[OptionGroup],
    ) -> None:
        super().__init__()
        self.module = module
        self.engine_driver = engine_driver
        self.client = client
        self.options = active_options
        self.all_options = self._collect_all_options(active_options)
        self.check_mode = self.module.check_mode
        self.param_cleanup: bool = self.module.params["cleanup"]
        self.param_container_default_behavior: t.Literal[
            "compatibility", "no_defaults"
        ] = self.module.params["container_default_behavior"]
        self.param_default_host_ip: str | None = self.module.params["default_host_ip"]
        self.param_debug: bool = self.module.params["debug"]
        self.param_force_kill: bool = self.module.params["force_kill"]
        self.param_image: str | None = self.module.params["image"]
        self.param_image_comparison: t.Literal["desired-image", "current-image"] = (
            self.module.params["image_comparison"]
        )
        self.param_image_label_mismatch: t.Literal["ignore", "fail"] = (
            self.module.params["image_label_mismatch"]
        )
        self.param_image_name_mismatch: t.Literal["ignore", "recreate"] = (
            self.module.params["image_name_mismatch"]
        )
        self.param_keep_volumes: bool = self.module.params["keep_volumes"]
        self.param_kill_signal: str | None = self.module.params["kill_signal"]
        self.param_name: str = self.module.params["name"]
        self.param_networks_cli_compatible: bool = self.module.params[
            "networks_cli_compatible"
        ]
        self.param_output_logs: bool = self.module.params["output_logs"]
        self.param_paused: bool | None = self.module.params["paused"]
        param_pull: t.Literal["never", "missing", "always", True, False] = (
            self.module.params["pull"]
        )
        if param_pull is True:
            param_pull = "always"
        if param_pull is False:
            param_pull = "missing"
        self.param_pull: t.Literal["never", "missing", "always"] = param_pull
        self.param_pull_check_mode_behavior: t.Literal[
            "image_not_present", "always"
        ] = self.module.params["pull_check_mode_behavior"]
        self.param_recreate: bool = self.module.params["recreate"]
        self.param_removal_wait_timeout: int | float | None = self.module.params[
            "removal_wait_timeout"
        ]
        self.param_healthy_wait_timeout: int | float | None = self.module.params[
            "healthy_wait_timeout"
        ]
        if (
            self.param_healthy_wait_timeout is not None
            and self.param_healthy_wait_timeout <= 0
        ):
            self.param_healthy_wait_timeout = None
        self.param_restart: bool = self.module.params["restart"]
        self.param_state: t.Literal[
            "absent", "present", "healthy", "started", "stopped"
        ] = self.module.params["state"]
        self._parse_comparisons()
        self._update_params()
        self.results = {"changed": False, "actions": []}
        self.diff: dict[str, t.Any] = {}
        self.diff_tracker = DifferenceTracker()
        self.facts: dict[str, t.Any] | None = {}
        if self.param_default_host_ip:
            valid_ip = False
            if re.match(
                r"^[0-9]+\.[0-9]+\.[0-9]+\.[0-9]+$", self.param_default_host_ip
            ):
                valid_ip = True
            if re.match(r"^\[[0-9a-fA-F:]+\]$", self.param_default_host_ip):
                valid_ip = True
            if re.match(r"^[0-9a-fA-F:]+$", self.param_default_host_ip):
                self.param_default_host_ip = f"[{self.param_default_host_ip}]"
                valid_ip = True
            if not valid_ip:
                self.fail(
                    "The value of default_host_ip must be an empty string, an IPv4 address, "
                    f'or an IPv6 address. Got "{self.param_default_host_ip}" instead.'
                )
        self.parameters: list[tuple[OptionGroup, dict[str, t.Any]]] | None = None

    def _add_action(self, action: dict[str, t.Any]) -> None:
        actions: list[dict[str, t.Any]] = self.results["actions"]  # type: ignore
        actions.append(action)

    def _collect_all_options(
        self, active_options: list[OptionGroup]
    ) -> dict[str, Option]:
        all_options = {}
        for options in active_options:
            for option in options.options:
                all_options[option.name] = option
        return all_options

    def _collect_all_module_params(self) -> set[str]:
        all_module_options = set()
        for option, data in self.module.argument_spec.items():
            all_module_options.add(option)
            if "aliases" in data:
                for alias in data["aliases"]:
                    all_module_options.add(alias)
        return all_module_options

    def _parse_comparisons(self) -> None:
        # Keep track of all module params and all option aliases
        all_module_options = self._collect_all_module_params()
        comp_aliases = {}
        for option_name, option in self.all_options.items():
            if option.not_an_ansible_option:
                continue
            comp_aliases[option_name] = option_name
            for alias in option.ansible_aliases:
                comp_aliases[alias] = option_name
        # Process comparisons specified by user
        comparisons: dict[str, t.Any] | None = self.module.params.get("comparisons")
        if comparisons:
            # If '*' appears in comparisons, process it first
            if "*" in comparisons:
                value = comparisons["*"]
                if value not in ("strict", "ignore"):
                    self.fail(
                        "The wildcard can only be used with comparison modes 'strict' and 'ignore'!"
                    )
                for option in self.all_options.values():
                    # `networks` is special: only update if
                    # some value is actually specified
                    if (
                        option.name == "networks"
                        and self.module.params["networks"] is None
                    ):
                        continue
                    option.comparison = value
            # Now process all other comparisons.
            comp_aliases_used: dict[str, str] = {}
            for key, value in comparisons.items():
                if key == "*":
                    continue
                # Find main key
                key_main = comp_aliases.get(key)
                if key_main is None:
                    if key_main in all_module_options:
                        self.fail(
                            f"The module option '{key}' cannot be specified in the comparisons dict, "
                            "since it does not correspond to container's state!"
                        )
                    if (
                        key not in self.all_options
                        or self.all_options[key].not_an_ansible_option
                    ):
                        self.fail(f"Unknown module option '{key}' in comparisons dict!")
                    key_main = key
                if key_main in comp_aliases_used:
                    self.fail(
                        f"Both '{key}' and '{comp_aliases_used[key_main]}' (aliases of {key_main}) are specified in comparisons dict!"
                    )
                comp_aliases_used[key_main] = key
                # Check value and update accordingly
                if value in ("strict", "ignore"):
                    self.all_options[key_main].comparison = value
                elif value == "allow_more_present":
                    if self.all_options[key_main].comparison_type == "value":
                        self.fail(
                            f"Option '{key}' is a value and not a set/list/dict, so its comparison cannot be {value}"
                        )
                    self.all_options[key_main].comparison = value
                else:
                    self.fail(f"Unknown comparison mode '{value}'!")
        # Copy values
        for option in self.all_options.values():
            if option.copy_comparison_from is not None:
                option.comparison = self.all_options[
                    option.copy_comparison_from
                ].comparison

    def _update_params(self) -> None:
        if (
            self.param_networks_cli_compatible is True
            and self.module.params["networks"]
            and self.module.params["network_mode"] is None
        ):
            # Same behavior as Docker CLI: if networks are specified, use the name of the first network as the value for network_mode
            # (assuming no explicit value is specified for network_mode)
            self.module.params["network_mode"] = self.module.params["networks"][0][
                "name"
            ]
        if self.param_container_default_behavior == "compatibility":
            old_default_values = {
                "auto_remove": False,
                "detach": True,
                "init": False,
                "interactive": False,
                "memory": "0",
                "paused": False,
                "privileged": False,
                "read_only": False,
                "tty": False,
            }
            for param, value in old_default_values.items():
                if self.module.params[param] is None:
                    self.module.params[param] = value

    def fail(self, *args: str, **kwargs: t.Any) -> t.NoReturn:
        # mypy doesn't know that Client has fail() method
        raise self.client.fail(*args, **kwargs)  # type: ignore

    def run(self) -> None:
        if self.param_state in ("stopped", "started", "present", "healthy"):
            # mypy doesn't get that self.param_state has only one of the above values
            self.present(self.param_state)  # type: ignore
        elif self.param_state == "absent":
            self.absent()

        if not self.check_mode and not self.param_debug:
            self.results.pop("actions")

        if self.module._diff or self.param_debug:
            self.diff["before"], self.diff["after"] = (
                self.diff_tracker.get_before_after()
            )
            self.results["diff"] = self.diff

        if self.facts:
            self.results["container"] = self.facts

    def wait_for_state(
        self,
        container_id: str,
        *,
        complete_states: Sequence[str | None] | None = None,
        wait_states: Sequence[str | None] | None = None,
        accept_removal: bool = False,
        max_wait: int | float | None = None,
        health_state: bool = False,
    ) -> dict[str, t.Any] | None:
        delay = 1.0
        total_wait = 0.0
        while True:
            # Inspect container
            result = self.engine_driver.inspect_container_by_id(
                self.client, container_id
            )
            if result is None:
                if accept_removal:
                    return result
                msg = f'Encontered vanished container while waiting for container "{container_id}"'
                self.fail(msg)
            # Check container state
            state_info = result.get("State") or {}
            if health_state:
                state_info = state_info.get("Health") or {}
            state = state_info.get("Status")
            if complete_states is not None and state in complete_states:
                return result
            if wait_states is not None and state not in wait_states:
                msg = f'Encontered unexpected state "{state}" while waiting for container "{container_id}"'
                self.fail(msg, container=result)
            # Wait
            if max_wait is not None:
                if total_wait > max_wait or delay < 1e-4:
                    msg = f'Timeout of {max_wait} seconds exceeded while waiting for container "{container_id}"'
                    self.fail(msg, container=result)
                if total_wait + delay > max_wait:
                    delay = max_wait - total_wait
            sleep(delay)
            total_wait += delay
            # Exponential backoff, but never wait longer than 10 seconds
            # (1.1**24 < 10, 1.1**25 > 10, so it will take 25 iterations
            #  until the maximal 10 seconds delay is reached. By then, the
            #  code will have slept for ~1.5 minutes.)
            delay = min(delay * 1.1, 10)

    def _collect_params(
        self, active_options: list[OptionGroup]
    ) -> list[tuple[OptionGroup, dict[str, t.Any]]]:
        parameters = []
        for options in active_options:
            values = {}
            engine = options.get_engine(self.engine_driver.name)
            for option in options.all_options:
                if (
                    not option.not_an_ansible_option
                    and self.module.params[option.name] is not None
                ):
                    values[option.name] = self.module.params[option.name]
            values = options.preprocess(self.module, values)
            engine.preprocess_value(
                self.module,
                self.client,
                self.engine_driver.get_api_version(self.client),
                options.options,
                values,
            )
            parameters.append((options, values))
        return parameters

    def _needs_container_image(self) -> bool:
        assert self.parameters is not None
        for options, values in self.parameters:
            engine = options.get_engine(self.engine_driver.name)
            if engine.needs_container_image(values):
                return True
        return False

    def _needs_host_info(self) -> bool:
        assert self.parameters is not None
        for options, values in self.parameters:
            engine = options.get_engine(self.engine_driver.name)
            if engine.needs_host_info(values):
                return True
        return False

    def present(
        self, state: t.Literal["stopped", "started", "present", "healthy"]
    ) -> None:
        self.parameters = self._collect_params(self.options)
        container = self._get_container(self.param_name)
        was_running = container.running
        was_paused = container.paused
        container_created = False

        # If the image parameter was passed then we need to deal with the image
        # version comparison. Otherwise we handle this depending on whether
        # the container already runs or not; in the former case, in case the
        # container needs to be restarted, we use the existing container's
        # image ID.
        image, container_image, comparison_image = self._get_image(
            container, needs_container_image=self._needs_container_image()
        )
        self.log(image, pretty_print=True)
        host_info = (
            self.engine_driver.get_host_info(self.client)
            if self._needs_host_info()
            else None
        )
        if not container.exists or container.removing:
            # New container
            if container.removing:
                self.log("Found container in removal phase")
            else:
                self.log("No container found")
            if not self.param_image:
                self.fail("Cannot create container when image is not specified!")
            self.diff_tracker.add("exists", parameter=True, active=False)
            if container.removing and not self.check_mode:
                # Wait for container to be removed before trying to create it
                assert container.id is not None
                self.wait_for_state(
                    container.id,
                    wait_states=["removing"],
                    accept_removal=True,
                    max_wait=self.param_removal_wait_timeout,
                )
            new_container = self.container_create(self.param_image)
            if new_container:
                container = new_container
            container_created = True
        else:
            # Existing container
            assert container.id is not None
            different, differences = self.has_different_configuration(
                container, container_image, comparison_image, host_info
            )
            image_different = False
            if self.all_options["image"].comparison == "strict":
                image_different = self._image_is_different(image, container)
                if (
                    self.param_image_name_mismatch == "recreate"
                    and self.param_image is not None
                    and self.param_image != container.image_name
                ):
                    different = True
                    self.diff_tracker.add(
                        "image_name",
                        parameter=self.param_image,
                        active=container.image_name,
                    )
            if image_different or different or self.param_recreate:
                self.diff_tracker.merge(differences)
                self.diff["differences"] = (
                    differences.get_legacy_docker_container_diffs()
                )
                if image_different:
                    self.diff["image_different"] = True
                self.log("differences")
                self.log(
                    differences.get_legacy_docker_container_diffs(), pretty_print=True
                )
                image_to_use = self.param_image
                if not image_to_use and container and container.image:
                    image_to_use = container.image
                if not image_to_use:
                    self.fail(
                        "Cannot recreate container when image is not specified or cannot be extracted from current container!"
                    )
                if container.running:
                    self.container_stop(container.id)
                self.container_remove(container.id)
                if not self.check_mode:
                    self.wait_for_state(
                        container.id,
                        wait_states=["removing"],
                        accept_removal=True,
                        max_wait=self.param_removal_wait_timeout,
                    )
                new_container = self.container_create(image_to_use)
                if new_container:
                    container = new_container
                container_created = True
                comparison_image = image

        if container and container.exists:
            container = self.update_limits(
                container, container_image, comparison_image, host_info
            )
            container = self.update_networks(container, container_created)

            if state in ("started", "healthy") and not container.running:
                self.diff_tracker.add("running", parameter=True, active=was_running)
                assert container.id is not None
                container = self.container_start(container.id)
            elif state in ("started", "healthy") and self.param_restart:
                self.diff_tracker.add("running", parameter=True, active=was_running)
                self.diff_tracker.add("restarted", parameter=True, active=False)
                assert container.id is not None
                container = self.container_restart(container.id)
            elif state == "stopped" and container.running:
                self.diff_tracker.add("running", parameter=False, active=was_running)
                assert container.id is not None
                self.container_stop(container.id)
                container = self._get_container(container.id)

            if (
                state in ("started", "healthy")
                and self.param_paused is not None
                and container.paused != self.param_paused
            ):
                self.diff_tracker.add(
                    "paused", parameter=self.param_paused, active=was_paused
                )
                if not self.check_mode:
                    assert container.id is not None
                    try:
                        if self.param_paused:
                            self.engine_driver.pause_container(
                                self.client, container.id
                            )
                        else:
                            self.engine_driver.unpause_container(
                                self.client, container.id
                            )
                    except Exception as exc:  # pylint: disable=broad-exception-caught
                        self.fail(
                            f"Error {'pausing' if self.param_paused else 'unpausing'} container {container.id}: {exc}"
                        )
                    container = self._get_container(container.id)
                self.results["changed"] = True
                self._add_action({"set_paused": self.param_paused})

        self.facts = container.raw

        if state == "healthy" and not self.check_mode:
            # `None` means that no health check enabled; simply treat this as 'healthy'
            assert container.id is not None
            inspect_result = self.wait_for_state(
                container.id,
                wait_states=["starting", "unhealthy"],
                complete_states=["healthy", None],
                max_wait=self.param_healthy_wait_timeout,
                health_state=True,
            )
            if inspect_result:
                # Return the latest inspection results retrieved
                self.facts = inspect_result

    def absent(self) -> None:
        container = self._get_container(self.param_name)
        if container.exists:
            assert container.id is not None
            if container.running:
                self.diff_tracker.add("running", parameter=False, active=True)
                self.container_stop(container.id)
            self.diff_tracker.add("exists", parameter=False, active=True)
            self.container_remove(container.id)

    def _output_logs(self, msg: str | bytes) -> None:
        self.module.log(msg=msg)

    def _get_container(self, container: str) -> Container:
        """
        Expects container ID or Name. Returns a container object
        """
        container_data = self.engine_driver.inspect_container_by_name(
            self.client, container
        )
        return Container(container_data, self.engine_driver)

    def _get_container_image(
        self, container: Container, fallback: dict[str, t.Any] | None = None
    ) -> dict[str, t.Any] | None:
        if not container.exists or container.removing:
            return fallback
        image = container.image
        assert image is not None
        if is_image_name_id(image):
            image_data = self.engine_driver.inspect_image_by_id(self.client, image)
        else:
            repository, tag = parse_repository_tag(image)
            if not tag:
                tag = "latest"
            image_data = self.engine_driver.inspect_image_by_name(
                self.client, repository, tag
            )
        return image_data or fallback

    def _get_image(
        self, container: Container, needs_container_image: bool = False
    ) -> tuple[
        dict[str, t.Any] | None, dict[str, t.Any] | None, dict[str, t.Any] | None
    ]:
        image_parameter = self.param_image
        get_container_image = needs_container_image or not image_parameter
        container_image = (
            self._get_container_image(container) if get_container_image else None
        )
        if container_image:
            self.log("current image")
            self.log(container_image, pretty_print=True)
        if not image_parameter:
            self.log("No image specified")
            return None, container_image, container_image
        if is_image_name_id(image_parameter):
            image = self.engine_driver.inspect_image_by_id(self.client, image_parameter)
            if image is None:
                self.fail(f"Cannot find image with ID {image_parameter}")
        else:
            repository, tag = parse_repository_tag(image_parameter)
            if not tag:
                tag = "latest"
            image = self.engine_driver.inspect_image_by_name(
                self.client, repository, tag
            )
            if not image and self.param_pull == "never":
                self.fail(
                    f"Cannot find image with name {repository}:{tag}, and pull=never"
                )
            if not image or self.param_pull == "always":
                if not self.check_mode:
                    self.log("Pull the image.")
                    image, already_to_latest = self.engine_driver.pull_image(
                        self.client,
                        repository,
                        tag,
                        image_platform=self.module.params["platform"],
                    )
                    if already_to_latest:
                        self.results["changed"] = False
                        self._add_action(
                            {"pulled_image": f"{repository}:{tag}", "changed": False}
                        )
                    else:
                        self.results["changed"] = True
                        self._add_action(
                            {"pulled_image": f"{repository}:{tag}", "changed": True}
                        )
                elif not image or self.param_pull_check_mode_behavior == "always":
                    # If the image is not there, or pull_check_mode_behavior == 'always', claim we'll
                    # pull. (Implicitly: if the image is there, claim it already was latest unless
                    # pull_check_mode_behavior == 'always'.)
                    self.results["changed"] = True
                    action: dict[str, t.Any] = {"pulled_image": f"{repository}:{tag}"}
                    if not image:
                        action["changed"] = True
                    self._add_action(action)

        self.log("image")
        self.log(image, pretty_print=True)

        comparison_image = image
        if self.param_image_comparison == "current-image":
            if not get_container_image:
                container_image = self._get_container_image(container)
            comparison_image = container_image

        return image, container_image, comparison_image

    def _image_is_different(
        self, image: dict[str, t.Any] | None, container: Container
    ) -> bool:
        if (
            image
            and image.get("Id")
            and container
            and container.image
            and image.get("Id") != container.image
        ):
            self.diff_tracker.add(
                "image", parameter=image.get("Id"), active=container.image
            )
            return True
        return False

    def _compose_create_parameters(self, image: str) -> dict[str, t.Any]:
        params: dict[str, t.Any] = {}
        assert self.parameters is not None
        for options, values in self.parameters:
            engine = options.get_engine(self.engine_driver.name)
            if engine.can_set_value(self.engine_driver.get_api_version(self.client)):
                engine.set_value(
                    self.module,
                    params,
                    self.engine_driver.get_api_version(self.client),
                    options.options,
                    values,
                )
        params["Image"] = image
        return params

    def _record_differences(
        self,
        differences: DifferenceTracker,
        options: OptionGroup,
        param_values: dict[str, t.Any],
        engine: Engine,
        container: Container,
        container_image: dict[str, t.Any] | None,
        image: dict[str, t.Any] | None,
        host_info: dict[str, t.Any] | None,
    ) -> None:
        assert container.raw is not None
        container_values = engine.get_value(
            self.module,
            container.raw,
            self.engine_driver.get_api_version(self.client),
            options.options,
            container_image,
            host_info,
        )
        expected_values = engine.get_expected_values(
            self.module,
            self.client,
            self.engine_driver.get_api_version(self.client),
            options.options,
            image,
            param_values.copy(),
            host_info,
        )
        for option in options.options:
            if option.name in expected_values:
                param_value = expected_values[option.name]
                container_value = container_values.get(option.name)
                match = engine.compare_value(option, param_value, container_value)

                if not match:
                    # No match.
                    if engine.ignore_mismatching_result(
                        self.module,
                        self.client,
                        self.engine_driver.get_api_version(self.client),
                        option,
                        image,
                        container_value,
                        param_value,
                        host_info,
                    ):
                        # Ignore the result
                        continue

                    # Record the differences
                    p = param_value
                    c = container_value
                    if option.comparison_type == "set":
                        # Since the order does not matter, sort so that the diff output is better.
                        if p is not None:
                            p = sorted(p)
                        if c is not None:
                            c = sorted(c)
                    elif option.comparison_type == "set(dict)":
                        # Since the order does not matter, sort so that the diff output is better.
                        if option.name == "expected_mounts":
                            # For selected values, use one entry as key
                            def sort_key_fn(x: dict[str, t.Any]) -> t.Any:
                                return x["target"]

                        else:
                            # We sort the list of dictionaries by using the sorted items of a dict as its key.
                            def sort_key_fn(x: dict[str, t.Any]) -> t.Any:
                                return sorted(
                                    (a, to_text(b, errors="surrogate_or_strict"))
                                    for a, b in x.items()
                                )

                        if p is not None:
                            p = sorted(p, key=sort_key_fn)
                        if c is not None:
                            c = sorted(c, key=sort_key_fn)
                    differences.add(option.name, parameter=p, active=c)

    def has_different_configuration(
        self,
        container: Container,
        container_image: dict[str, t.Any] | None,
        image: dict[str, t.Any] | None,
        host_info: dict[str, t.Any] | None,
    ) -> tuple[bool, DifferenceTracker]:
        differences = DifferenceTracker()
        update_differences = DifferenceTracker()
        assert self.parameters is not None
        for options, param_values in self.parameters:
            engine = options.get_engine(self.engine_driver.name)
            if engine.can_update_value(self.engine_driver.get_api_version(self.client)):
                self._record_differences(
                    update_differences,
                    options,
                    param_values,
                    engine,
                    container,
                    container_image,
                    image,
                    host_info,
                )
            else:
                self._record_differences(
                    differences,
                    options,
                    param_values,
                    engine,
                    container,
                    container_image,
                    image,
                    host_info,
                )
        has_differences = not differences.empty
        # Only consider differences of properties that can be updated when there are also other differences
        if has_differences:
            differences.merge(update_differences)
        return has_differences, differences

    def has_different_resource_limits(
        self,
        container: Container,
        container_image: dict[str, t.Any] | None,
        image: dict[str, t.Any] | None,
        host_info: dict[str, t.Any] | None,
    ) -> tuple[bool, DifferenceTracker]:
        differences = DifferenceTracker()
        assert self.parameters is not None
        for options, param_values in self.parameters:
            engine = options.get_engine(self.engine_driver.name)
            if not engine.can_update_value(
                self.engine_driver.get_api_version(self.client)
            ):
                continue
            self._record_differences(
                differences,
                options,
                param_values,
                engine,
                container,
                container_image,
                image,
                host_info,
            )
        has_differences = not differences.empty
        return has_differences, differences

    def _compose_update_parameters(self) -> dict[str, t.Any]:
        result: dict[str, t.Any] = {}
        assert self.parameters is not None
        for options, values in self.parameters:
            engine = options.get_engine(self.engine_driver.name)
            if not engine.can_update_value(
                self.engine_driver.get_api_version(self.client)
            ):
                continue
            engine.update_value(
                self.module,
                result,
                self.engine_driver.get_api_version(self.client),
                options.options,
                values,
            )
        return result

    def update_limits(
        self,
        container: Container,
        container_image: dict[str, t.Any] | None,
        image: dict[str, t.Any] | None,
        host_info: dict[str, t.Any] | None,
    ) -> Container:
        limits_differ, different_limits = self.has_different_resource_limits(
            container, container_image, image, host_info
        )
        if limits_differ:
            self.log("limit differences:")
            self.log(
                different_limits.get_legacy_docker_container_diffs(), pretty_print=True
            )
            self.diff_tracker.merge(different_limits)
        if limits_differ and not self.check_mode:
            assert container.id is not None
            self.container_update(container.id, self._compose_update_parameters())
            return self._get_container(container.id)
        return container

    def has_network_differences(
        self, container: Container
    ) -> tuple[bool, list[dict[str, t.Any]]]:
        """
        Check if the container is connected to requested networks with expected options: links, aliases, ipv4, ipv6
        """
        different = False
        differences: list[dict[str, t.Any]] = []

        if not self.module.params["networks"]:
            return different, differences

        assert container.container is not None
        if not container.container.get("NetworkSettings"):
            self.fail(
                "has_missing_networks: Error parsing container properties. NetworkSettings missing."
            )

        connected_networks = container.container["NetworkSettings"]["Networks"]
        for network in self.module.params["networks"]:
            network_info = connected_networks.get(network["name"])
            if network_info is None:
                different = True
                differences.append({"parameter": network, "container": None})
            else:
                diff = False
                network_info_ipam = network_info.get("IPAMConfig") or {}
                if network.get("ipv4_address") and normalize_ip_address(
                    network["ipv4_address"]
                ) != normalize_ip_address(network_info_ipam.get("IPv4Address")):
                    diff = True
                if network.get("ipv6_address") and normalize_ip_address(
                    network["ipv6_address"]
                ) != normalize_ip_address(network_info_ipam.get("IPv6Address")):
                    diff = True
                if network.get("aliases") and not compare_generic(
                    network["aliases"],
                    network_info.get("Aliases"),
                    "allow_more_present",
                    "set",
                ):
                    diff = True
                if network.get("links"):
                    expected_links = []
                    for link, alias in network["links"]:
                        expected_links.append(f"{link}:{alias}")
                    if not compare_generic(
                        expected_links,
                        network_info.get("Links"),
                        "allow_more_present",
                        "set",
                    ):
                        diff = True
                if network.get("mac_address") and network[
                    "mac_address"
                ] != network_info.get("MacAddress"):
                    diff = True
                if diff:
                    different = True
                    differences.append(
                        {
                            "parameter": network,
                            "container": {
                                "name": network["name"],
                                "ipv4_address": network_info_ipam.get("IPv4Address"),
                                "ipv6_address": network_info_ipam.get("IPv6Address"),
                                "aliases": network_info.get("Aliases"),
                                "links": network_info.get("Links"),
                                "mac_address": network_info.get("MacAddress"),
                            },
                        }
                    )
        return different, differences

    def has_extra_networks(
        self, container: Container
    ) -> tuple[bool, list[dict[str, t.Any]]]:
        """
        Check if the container is connected to non-requested networks
        """
        extra_networks: list[dict[str, t.Any]] = []
        extra = False

        assert container.container is not None
        if not container.container.get("NetworkSettings"):
            self.fail(
                "has_extra_networks: Error parsing container properties. NetworkSettings missing."
            )

        connected_networks = container.container["NetworkSettings"].get("Networks")
        if connected_networks:
            for network, network_config in connected_networks.items():
                keep = False
                if self.module.params["networks"]:
                    for expected_network in self.module.params["networks"]:
                        if expected_network["name"] == network:
                            keep = True
                if not keep:
                    extra = True
                    extra_networks.append(
                        {"name": network, "id": network_config["NetworkID"]}
                    )
        return extra, extra_networks

    def update_networks(
        self, container: Container, container_created: bool
    ) -> Container:
        updated_container = container
        if self.all_options["networks"].comparison != "ignore" or container_created:
            has_network_differences, network_differences = self.has_network_differences(
                container
            )
            if has_network_differences:
                if self.diff.get("differences"):
                    self.diff["differences"].append(
                        {"network_differences": network_differences}
                    )
                else:
                    self.diff["differences"] = [
                        {"network_differences": network_differences}
                    ]
                for netdiff in network_differences:
                    self.diff_tracker.add(
                        f"network.{netdiff['parameter']['name']}",
                        parameter=netdiff["parameter"],
                        active=netdiff["container"],
                    )
                self.results["changed"] = True
                updated_container = self._add_networks(container, network_differences)

        purge_networks = (
            self.all_options["networks"].comparison == "strict"
            and self.module.params["networks"] is not None
        )
        if purge_networks:
            has_extra_networks, extra_networks = self.has_extra_networks(container)
            if has_extra_networks:
                if self.diff.get("differences"):
                    self.diff["differences"].append({"purge_networks": extra_networks})
                else:
                    self.diff["differences"] = [{"purge_networks": extra_networks}]
                for extra_network in extra_networks:
                    self.diff_tracker.add(
                        f"network.{extra_network['name']}", active=extra_network
                    )
                self.results["changed"] = True
                updated_container = self._purge_networks(container, extra_networks)
        return updated_container

    def _add_networks(
        self, container: Container, differences: list[dict[str, t.Any]]
    ) -> Container:
        assert container.id is not None
        for diff in differences:
            # remove the container from the network, if connected
            if diff.get("container"):
                self._add_action({"removed_from_network": diff["parameter"]["name"]})
                if not self.check_mode:
                    try:
                        self.engine_driver.disconnect_container_from_network(
                            self.client, container.id, diff["parameter"]["id"]
                        )
                    except Exception as exc:  # pylint: disable=broad-exception-caught
                        self.fail(
                            f"Error disconnecting container from network {diff['parameter']['name']} - {exc}"
                        )
            # connect to the network
            self._add_action(
                {
                    "added_to_network": diff["parameter"]["name"],
                    "network_parameters": diff["parameter"],
                }
            )
            if not self.check_mode:
                params = {
                    key: value
                    for key, value in diff["parameter"].items()
                    if key not in ("id", "name")
                }
                try:
                    self.log(
                        f"Connecting container to network {diff['parameter']['id']}"
                    )
                    self.log(params, pretty_print=True)
                    self.engine_driver.connect_container_to_network(
                        self.client, container.id, diff["parameter"]["id"], params
                    )
                except Exception as exc:  # pylint: disable=broad-exception-caught
                    self.fail(
                        f"Error connecting container to network {diff['parameter']['name']} - {exc}"
                    )
        return self._get_container(container.id)

    def _purge_networks(
        self, container: Container, networks: list[dict[str, t.Any]]
    ) -> Container:
        assert container.id is not None
        for network in networks:
            self._add_action({"removed_from_network": network["name"]})
            if not self.check_mode:
                try:
                    self.engine_driver.disconnect_container_from_network(
                        self.client, container.id, network["name"]
                    )
                except Exception as exc:  # pylint: disable=broad-exception-caught
                    self.fail(
                        f"Error disconnecting container from network {network['name']} - {exc}"
                    )
        return self._get_container(container.id)

    def container_create(self, image: str) -> Container | None:
        create_parameters = self._compose_create_parameters(image)
        self.log("create container")
        self.log(f"image: {image} parameters:")
        self.log(create_parameters, pretty_print=True)
        networks = {}
        if self.param_networks_cli_compatible and self.module.params["networks"]:
            network_list = self.module.params["networks"]
            if not self.engine_driver.create_container_supports_more_than_one_network(
                self.client
            ):
                network_list = network_list[:1]
            for network in network_list:
                networks[network["name"]] = {
                    key: value
                    for key, value in network.items()
                    if key not in ("name", "id")
                }
        self._add_action(
            {
                "created": "Created container",
                "create_parameters": create_parameters,
                "networks": networks,
            }
        )
        self.results["changed"] = True
        if not self.check_mode:
            try:
                container_id = self.engine_driver.create_container(
                    self.client, self.param_name, create_parameters, networks=networks
                )
            except Exception as exc:  # pylint: disable=broad-exception-caught
                self.fail(f"Error creating container: {exc}")
            return self._get_container(container_id)
        return None

    def container_start(self, container_id: str) -> Container:
        self.log(f"start container {container_id}")
        self._add_action({"started": container_id})
        self.results["changed"] = True
        if not self.check_mode:
            try:
                self.engine_driver.start_container(self.client, container_id)
            except Exception as exc:  # pylint: disable=broad-exception-caught
                self.fail(f"Error starting container {container_id}: {exc}")

            if self.module.params["detach"] is False:
                status = self.engine_driver.wait_for_container(
                    self.client, container_id
                )
                # mypy doesn't know that Client has fail_results property
                self.client.fail_results["status"] = status  # type: ignore
                self.results["status"] = status

                output: str | bytes
                if self.module.params["auto_remove"]:
                    output = "Cannot retrieve result as auto_remove is enabled"
                    if self.param_output_logs:
                        self.module.warn(
                            "Cannot output_logs if auto_remove is enabled!"
                        )
                else:
                    output, real_output = self.engine_driver.get_container_output(
                        self.client, container_id
                    )
                    if real_output and self.param_output_logs:
                        self._output_logs(msg=output)

                if self.param_cleanup:
                    self.container_remove(container_id, force=True)
                insp = self._get_container(container_id)
                if insp.raw:
                    insp.raw["Output"] = output
                else:
                    insp.raw = {"Output": output}
                if status != 0:
                    # Set `failed` to True and return output as msg
                    self.results["failed"] = True
                    self.results["msg"] = output
                return insp
        return self._get_container(container_id)

    def container_remove(
        self, container_id: str, link: bool = False, force: bool = False
    ) -> None:
        volume_state = not self.param_keep_volumes
        self.log(
            f"remove container container:{container_id} v:{volume_state} link:{link} force{force}"
        )
        self._add_action(
            {
                "removed": container_id,
                "volume_state": volume_state,
                "link": link,
                "force": force,
            }
        )
        self.results["changed"] = True
        if not self.check_mode:
            try:
                self.engine_driver.remove_container(
                    self.client,
                    container_id,
                    remove_volumes=volume_state,
                    link=link,
                    force=force,
                )
            except Exception as exc:  # pylint: disable=broad-exception-caught
                self.fail(f"Error removing container {container_id}: {exc}")

    def container_update(
        self, container_id: str, update_parameters: dict[str, t.Any]
    ) -> Container:
        if update_parameters:
            self.log(f"update container {container_id}")
            self.log(update_parameters, pretty_print=True)
            self._add_action(
                {"updated": container_id, "update_parameters": update_parameters}
            )
            self.results["changed"] = True
            if not self.check_mode:
                try:
                    self.engine_driver.update_container(
                        self.client, container_id, update_parameters
                    )
                except Exception as exc:  # pylint: disable=broad-exception-caught
                    self.fail(f"Error updating container {container_id}: {exc}")
        return self._get_container(container_id)

    def container_kill(self, container_id: str) -> None:
        self._add_action({"killed": container_id, "signal": self.param_kill_signal})
        self.results["changed"] = True
        if not self.check_mode:
            try:
                self.engine_driver.kill_container(
                    self.client, container_id, kill_signal=self.param_kill_signal
                )
            except Exception as exc:  # pylint: disable=broad-exception-caught
                self.fail(f"Error killing container {container_id}: {exc}")

    def container_restart(self, container_id: str) -> Container:
        self._add_action(
            {"restarted": container_id, "timeout": self.module.params["stop_timeout"]}
        )
        self.results["changed"] = True
        if not self.check_mode:
            try:
                self.engine_driver.restart_container(
                    self.client, container_id, self.module.params["stop_timeout"] or 10
                )
            except Exception as exc:  # pylint: disable=broad-exception-caught
                self.fail(f"Error restarting container {container_id}: {exc}")
        return self._get_container(container_id)

    def container_stop(self, container_id: str) -> None:
        if self.param_force_kill:
            self.container_kill(container_id)
            return
        self._add_action(
            {"stopped": container_id, "timeout": self.module.params["stop_timeout"]}
        )
        self.results["changed"] = True
        if not self.check_mode:
            try:
                self.engine_driver.stop_container(
                    self.client, container_id, self.module.params["stop_timeout"]
                )
            except Exception as exc:  # pylint: disable=broad-exception-caught
                self.fail(f"Error stopping container {container_id}: {exc}")


def run_module(engine_driver: EngineDriver) -> None:
    module, active_options, client = engine_driver.setup(
        argument_spec={
            "cleanup": {"type": "bool", "default": False},
            "comparisons": {"type": "dict"},
            "container_default_behavior": {
                "type": "str",
                "default": "no_defaults",
                "choices": ["compatibility", "no_defaults"],
            },
            "command_handling": {
                "type": "str",
                "choices": ["compatibility", "correct"],
                "default": "correct",
            },
            "default_host_ip": {"type": "str"},
            "force_kill": {"type": "bool", "default": False, "aliases": ["forcekill"]},
            "image": {"type": "str"},
            "image_comparison": {
                "type": "str",
                "choices": ["desired-image", "current-image"],
                "default": "desired-image",
            },
            "image_label_mismatch": {
                "type": "str",
                "choices": ["ignore", "fail"],
                "default": "ignore",
            },
            "image_name_mismatch": {
                "type": "str",
                "choices": ["ignore", "recreate"],
                "default": "recreate",
            },
            "keep_volumes": {"type": "bool", "default": True},
            "kill_signal": {"type": "str"},
            "name": {"type": "str", "required": True},
            "networks_cli_compatible": {"type": "bool", "default": True},
            "output_logs": {"type": "bool", "default": False},
            "paused": {"type": "bool"},
            "pull": {
                "type": "raw",
                "choices": ["never", "missing", "always", True, False],
                "default": "missing",
            },
            "pull_check_mode_behavior": {
                "type": "str",
                "choices": ["image_not_present", "always"],
                "default": "image_not_present",
            },
            "recreate": {"type": "bool", "default": False},
            "removal_wait_timeout": {"type": "float"},
            "restart": {"type": "bool", "default": False},
            "state": {
                "type": "str",
                "default": "started",
                "choices": ["absent", "present", "healthy", "started", "stopped"],
            },
            "healthy_wait_timeout": {"type": "float", "default": 300},
        },
        required_if=[
            ("state", "present", ["image"]),
        ],
    )

    def execute() -> t.NoReturn:
        cm = ContainerManager(module, engine_driver, client, active_options)
        cm.run()
        module.exit_json(**sanitize_result(cm.results))

    engine_driver.run(execute, client)



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Copyright (c) 2021, Cisco Systems
# GNU General Public License v3.0+ (see LICENSE or
# https://www.gnu.org/licenses/gpl-3.0.txt)

DOCUMENTATION = r"""
module: networks_appliance_vpn_bgp
short_description: Resource module for networks _appliance _vpn _bgp
description:
  - Manage operation update of the resource networks _appliance _vpn _bgp.
  - Update a Hub BGP Configuration.
version_added: '1.0.0'
extends_documentation_fragment:
  - cisco.meraki.module
author: Francisco Munoz (@fmunoz)
options:
  asNumber:
    description: An Autonomous System Number (ASN) is required if you are to run BGP
      and peer with another BGP Speaker outside of the Auto VPN domain. This ASN will
      be applied to the entire Auto VPN domain. The entire 4-byte ASN range is supported.
      So, the ASN must be an integer between 1 and 4294967295. When absent, this field
      is not updated. If no value exists then it defaults to 64512.
    type: int
  enabled:
    description: Boolean value to enable or disable the BGP configuration. When BGP
      is enabled, the asNumber (ASN) will be autopopulated with the preconfigured
      ASN at other Hubs or a default value if there is no ASN configured.
    type: bool
  ibgpHoldTimer:
    description: The iBGP holdtimer in seconds. The iBGP holdtimer must be an integer
      between 12 and 240. When absent, this field is not updated. If no value exists
      then it defaults to 240.
    type: int
  neighbors:
    description: List of BGP neighbors. This list replaces the existing set of neighbors.
      When absent, this field is not updated.
    elements: dict
    suboptions:
      allowTransit:
        description: When this feature is on, the Meraki device will advertise routes
          learned from other Autonomous Systems, thereby allowing traffic between
          Autonomous Systems to transit this AS. When absent, it defaults to false.
        type: bool
      authentication:
        description: Authentication settings between BGP peers.
        suboptions:
          password:
            description: Password to configure MD5 authentication between BGP peers.
            type: str
        type: dict
      communityOut:
        description: List of BGP communities tagged to the routes advertised to an
          eBGP neighbor.
        elements: str
        type: list
      ebgpHoldTimer:
        description: The eBGP hold timer in seconds for each neighbor. The eBGP hold
          timer must be an integer between 12 and 240.
        type: int
      ebgpMultihop:
        description: Configure this if the neighbor is not adjacent. The eBGP multi-hop
          must be an integer between 1 and 255.
        type: int
      filterIn:
        description: Filters routes received from an eBGP neighbor. Each entry must
          be an IPv4 CIDR string (e.g., '10.0.0.0/8' or '10.0.0.5/32').
        elements: str
        type: list
      ip:
        description: The IPv4 address of the neighbor.
        type: str
      ipv6:
        description: Information regarding IPv6 address of the neighbor, Required
          if `ip` is not present.
        suboptions:
          address:
            description: The IPv6 address of the neighbor.
            type: str
        type: dict
      multiExitDiscriminator:
        description: Configures the local metric associated with routes received from
          the remote peer. Routes from peers with lower metrics are will be preferred.
          Must be an integer between 0 and 4294967295. MED is 6th in the decision
          tree when identical routes from multiple peers exist.
        type: int
      nextHopIp:
        description: The IPv4 address of the remote BGP peer that will establish a
          TCP session with the local MX.
        type: str
      pathPrepend:
        description: Prepends the AS_PATH BGP Attribute associated with routes received
          from the remote peer. Configurable value of ASNs to prepend. Length of the
          array may not exceed 10, and each ASN in the array must be an integer between
          1 and 4294967295. AS_PATH is 4th in the decision tree when identical routes
          from multiple peers exist.
        elements: int
        type: list
      receiveLimit:
        description: The receive limit is the maximum number of routes that can be
          received from any BGP peer. The receive limit must be an integer between
          0 and 2147483647. When absent, it defaults to 0.
        type: int
      remoteAsNumber:
        description: Remote ASN of the neighbor. The remote ASN must be an integer
          between 1 and 4294967295.
        type: int
      sourceInterface:
        description: The output interface for peering with the remote BGP peer. Valid
          values are 'wan{NUMBER}' (e.g. 'wan3') or 'vlan{VLAN ID}' (e.g. 'vlan123')
          for MX; 'l3lan{ID}' (e.g. 'l3lan123') for Secure Router L3 interfaces.
        type: str
      ttlSecurity:
        description: Settings for BGP TTL security to protect BGP peering sessions
          from forged IP attacks.
        suboptions:
          enabled:
            description: Boolean value to enable or disable BGP TTL security.
            type: bool
        type: dict
      weight:
        description: Sets the local weight for routes received from the remote peer.
          Routes from peers with higher weights will be preferred. Must be an integer
          between 0 and 49.
        type: int
    type: list
  networkId:
    description: NetworkId path parameter. Network ID.
    type: str
  routerId:
    description: The router ID of the appliance.
    type: str
requirements:
  - meraki >= 2.4.9
  - python >= 3.5
seealso:
  - name: Cisco Meraki documentation for appliance updateNetworkApplianceVpnBgp
    description: Complete reference of the updateNetworkApplianceVpnBgp API.
    link: https://developer.cisco.com/meraki/api-v1/#!update-network-appliance-vpn-bgp
notes:
  - SDK Method used are
    appliance.Appliance.update_network_appliance_vpn_bgp,
  - Paths used are
    put /networks/{networkId}/appliance/vpn/bgp,
"""

EXAMPLES = r"""
- name: Update all
  cisco.meraki.networks_appliance_vpn_bgp:
    meraki_api_key: "{{ meraki_api_key }}"
    meraki_base_url: "{{ meraki_base_url }}"
    meraki_single_request_timeout: "{{ meraki_single_request_timeout }}"
    meraki_certificate_path: "{{ meraki_certificate_path }}"
    meraki_requests_proxy: "{{ meraki_requests_proxy }}"
    meraki_wait_on_rate_limit: "{{ meraki_wait_on_rate_limit }}"
    meraki_nginx_429_retry_wait_time: "{{ meraki_nginx_429_retry_wait_time }}"
    meraki_action_batch_retry_wait_time: "{{ meraki_action_batch_retry_wait_time }}"
    meraki_retry_4xx_error: "{{ meraki_retry_4xx_error }}"
    meraki_retry_4xx_error_wait_time: "{{ meraki_retry_4xx_error_wait_time }}"
    meraki_maximum_retries: "{{ meraki_maximum_retries }}"
    meraki_output_log: "{{ meraki_output_log }}"
    meraki_log_file_prefix: "{{ meraki_log_file_prefix }}"
    meraki_log_path: "{{ meraki_log_path }}"
    meraki_print_console: "{{ meraki_print_console }}"
    meraki_suppress_logging: "{{ meraki_suppress_logging }}"
    meraki_simulate: "{{ meraki_simulate }}"
    meraki_be_geo_id: "{{ meraki_be_geo_id }}"
    meraki_caller: "{{ meraki_caller }}"
    meraki_use_iterator_for_get_pages: "{{ meraki_use_iterator_for_get_pages }}"
    meraki_inherit_logging_config: "{{ meraki_inherit_logging_config }}"
    state: present
    asNumber: 64515
    enabled: true
    ibgpHoldTimer: 120
    neighbors:
      - allowTransit: true
        authentication:
          password: abc123
        communityOut:
          - 64515:100
          - NO_EXPORT
        ebgpHoldTimer: 180
        ebgpMultihop: 2
        filterIn:
          - 10.0.0.0/8
          - 172.16.0.0/12
        ip: 10.10.10.22
        ipv6:
          address: 2002::1234:abcd:ffff:c0a8:101
        multiExitDiscriminator: 2
        nextHopIp: 1.2.3.4
        pathPrepend:
          - 1
          - 2
        receiveLimit: 120
        remoteAsNumber: 64343
        sourceInterface: wan1
        ttlSecurity:
          enabled: false
        weight: 10
    networkId: string
    routerId: 10.15.10.2
"""
RETURN = r"""
meraki_response:
  description: A dictionary or list with the response returned by the Cisco Meraki Python SDK
  returned: always
  type: dict
  sample: >
    {
      "enabled": true,
      "asNumber": 0,
      "ibgpHoldTimer": 0,
      "routerId": "string",
      "neighbors": [
        {
          "ip": "string",
          "ipv6": {
            "address": "string"
          },
          "remoteAsNumber": 0,
          "receiveLimit": 0,
          "allowTransit": true,
          "ebgpHoldTimer": 0,
          "ebgpMultihop": 0,
          "sourceInterface": "string",
          "nextHopIp": "string",
          "ttlSecurity": {
            "enabled": true
          },
          "authentication": {
            "password": "string"
          },
          "multiExitDiscriminator": 0,
          "pathPrepend": [
            0
          ],
          "weight": 0,
          "filterIn": [
            "string"
          ],
          "communityOut": [
            "string"
          ]
        }
      ]
    }
"""



#!/usr/bin/python
#
# Copyright (c) 2020 Fred-Sun, (@Fred-Sun)
#
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import absolute_import, division, print_function
__metaclass__ = type


DOCUMENTATION = '''
---
module: azure_rm_virtualwan
version_added: '1.5.0'
short_description: Manage Azure VirtualWan instance
description:
    - Create, update and delete instance of Azure VirtualWan.
options:
    resource_group:
        description:
            - The resource group name of the VirtualWan.
        required: true
        type: str
    office365_local_breakout_category:
        description:
            - Specifies the Office365 local breakout category.
            - Default value is C(None).
        type: str
        choices:
            - Optimize
            - OptimizeAndAllow
            - All
            - None
    name:
        description:
            - The name of the VirtualWAN being retrieved.
        required: true
        type: str
    location:
        description:
            - The virtual wan location.
        type: str
    disable_vpn_encryption:
        description:
            - Vpn encryption to be disabled or not.
        type: bool
    virtual_hubs:
        description:
            - List of VirtualHubs in the VirtualWAN.
        type: list
        elements: dict
        suboptions:
            id:
                description:
                    - The virtual hub resource ID.
                type: str
    vpn_sites:
        description:
            - List of VpnSites in the VirtualWAN.
        type: list
        elements: dict
        suboptions:
            id:
                description:
                    - The vpn site resource ID.
                type: str
    allow_branch_to_branch_traffic:
        description:
            - True if branch to branch traffic is allowed.
        type: bool
    allow_vnet_to_vnet_traffic:
        description:
            - C(True) if Vnet to Vnet traffic is allowed.
        type: bool
    virtual_wan_type:
        description:
            - The type of the VirtualWAN.
        type: str
        choices:
            - Basic
            - Standard
    state:
        description:
            - Assert the state of the VirtualWan.
            - Use C(present) to create or update an VirtualWan and C(absent) to delete it.
        default: present
        type: str
        choices:
            - absent
            - present
extends_documentation_fragment:
    - azure.azcollection.azure
    - azure.azcollection.azure_tags
author:
    - Fred-Sun (@Fred-Sun)

'''

EXAMPLES = '''
- name: Create a VirtualWan
  azure_rm_virtualwan:
    resource_group: myResouceGroup
    name: testwan
    disable_vpn_encryption: true
    allow_branch_to_branch_traffic: true
    allow_vnet_to_vnet_traffic: true
    virtual_wan_type: Standard

- name: Delete the VirtualWan
  azure_rm_virtualwan:
    resource_group: myResouceGroup
    name: testwan
    state: absent
'''

RETURN = '''
state:
    description:
        - Current state of the virtual wan.
    type: complex
    returned: success
    contains:
        id:
            description:
                - Resource ID.
            returned: always
            type: str
            sample: /subscriptions/xxx-xxx/resourceGroups/myResourceGroup/providers/Microsoft.Network/virtualWans/virtual_wan_name
        name:
            description:
                - Resource name.
            returned: always
            type: str
            sample: virtualwanb57dc9555691
        type:
            description:
                - Resource type.
            returned: always
            type: str
            sample: Microsoft.Network/virtualWans
        location:
            description:
                - The virtual wan resource location.
            returned: always
            type: str
            sample: eastus
        tags:
            description:
                - Resource tags.
            returned: always
            type: dict
            sample: { 'key1': 'value1'}
        etag:
            description:
                - A unique read-only string that changes whenever the resource is updated.
            returned: always
            type: str
            sample: 52def36b-84b6-49aa-a825-16ba167fc559
        disable_vpn_encryption:
            description:
                - Vpn encryption to be disabled or not.
            returned: always
            type: bool
            sample: true
        virtual_hubs:
            description:
                - List of VirtualHubs in the VirtualWAN.
            type: complex
            returned: always
            contains:
                id:
                    description:
                        - The virtual hubs ID.
                    type: str
                    returned: always
                    sample: /subscriptions/xxx-xxx/resourceGroups/myResourceGroup/providers/Microsoft.Network/virtualHubs/test
        vpn_sites:
            description:
                - List of VpnSites in the VirtualWAN.
            returned: always
            type: list
            contains:
                id:
                    description:
                        - The vpn sites resouce ID.
                    returned: always
                    type: str
                    sample: /subscriptions/xxx-xxx/resourceGroups/resource_group/providers/Microsoft.Network/vpnSites/test1
        allow_branch_to_branch_traffic:
            description:
                - True if branch to branch traffic is allowed.
            returned: always
            type: bool
            sample: true
        allow_vnet_to_vnet_traffic:
            description:
                - True if Vnet to Vnet traffic is allowed.
            returned: always
            type: bool
            sample: true
        office365_local_breakout_category:
            description:
                - The office local breakout category.
            returned: always
            type: str
            sample: None
        provisioning_state:
            description:
                - The provisioning state of the virtual WAN resource.
            returned: always
            type: str
            sample: Succeeded
        virtual_wan_type:
            description:
                - The type of the VirtualWAN.
            returned: always
            type: str
            sample: Standard

'''

from ansible_collections.azure.azcollection.plugins.module_utils.azure_rm_common_ext import AzureRMModuleBaseExt
try:
    from azure.core.exceptions import ResourceNotFoundError
    from azure.core.polling import LROPoller
except ImportError:
    # This is handled in azure_rm_common
    pass


class Actions:
    NoAction, Create, Update, Update_tags, Delete = range(5)


class AzureRMVirtualWan(AzureRMModuleBaseExt):
    def __init__(self):
        self.module_arg_spec = dict(
            location=dict(
                type='str'
            ),
            resource_group=dict(
                type='str',
                required=True
            ),
            office365_local_breakout_category=dict(
                type='str',
                choices=['Optimize', 'OptimizeAndAllow', 'All', 'None']
            ),
            name=dict(
                type='str',
                required=True
            ),
            disable_vpn_encryption=dict(
                type='bool'
            ),
            virtual_hubs=dict(
                type='list',
                elements='dict',
                options=dict(
                    id=dict(
                        type='str'
                    )
                )
            ),
            vpn_sites=dict(
                type='list',
                elements='dict',
                options=dict(
                    id=dict(
                        type='str'
                    )
                )
            ),
            allow_branch_to_branch_traffic=dict(
                type='bool'
            ),
            allow_vnet_to_vnet_traffic=dict(
                type='bool'
            ),
            virtual_wan_type=dict(
                type='str',
                choices=['Basic', 'Standard']
            ),
            state=dict(
                type='str',
                default='present',
                choices=['present', 'absent']
            )
        )

        self.resource_group = None
        self.name = None
        self.location = None
        self.body = {}

        self.results = dict(changed=False)
        self.state = None
        self.to_do = Actions.NoAction

        super(AzureRMVirtualWan, self).__init__(derived_arg_spec=self.module_arg_spec,
                                                supports_check_mode=True,
                                                supports_tags=True)

    def exec_module(self, **kwargs):
        for key in list(self.module_arg_spec.keys()) + ['tags']:
            if hasattr(self, key):
                setattr(self, key, kwargs[key])
            elif kwargs[key] is not None:
                self.body[key] = kwargs[key]

        resource_group = self.get_resource_group(self.resource_group)
        if self.location is None:
            # Set default location
            self.location = resource_group.location
        self.body['location'] = self.location

        old_response = None
        response = None

        old_response = self.get_resource()

        if not old_response:
            if self.state == 'present':
                self.to_do = Actions.Create
        else:
            if self.state == 'absent':
                self.to_do = Actions.Delete
            else:
                compare_list = ['disable_vpn_encryption', 'allow_branch_to_branch_traffic']
                for key in compare_list:
                    if self.body.get(key) is not None and self.body[key] != old_response[key]:
                        self.log('parameter {0} does not match the configuration'.format(key))
                        self.to_do = Actions.Update
                    else:
                        self.body[key] = old_response[key]

                update_tags, self.tags = self.update_tags(old_response.get('tags'))
                if update_tags:
                    self.to_do = Actions.Update_tags

        if (self.to_do == Actions.Create) or (self.to_do == Actions.Update):
            self.results['changed'] = True
            if self.check_mode:
                return self.results
            response = self.create_update_resource()
        elif self.to_do == Actions.Update_tags:
            self.results['changed'] = True
            if self.check_mode:
                return self.results
            response = self.update_resource_tags(dict(tags=self.tags))
        elif self.to_do == Actions.Delete:
            self.results['changed'] = True
            if self.check_mode:
                return self.results
            self.delete_resource()
        else:
            self.results['changed'] = False
            response = old_response
        if response is not None:
            self.results['state'] = response
        return self.results

    def update_resource_tags(self, tags_parameters):
        try:
            response = self.network_client.virtual_wans.update_tags(resource_group_name=self.resource_group,
                                                                    virtual_wan_name=self.name,
                                                                    wan_parameters=tags_parameters)
            if isinstance(response, LROPoller):
                response = self.get_poller_result(response)
        except Exception as exc:
            self.log('Error attempting to Update the VirtualWan instance.')
            self.fail('Error Updating the VirtualWan instance tags: {0}'.format(str(exc)))
        return response.as_dict()

    def create_update_resource(self):
        try:
            response = self.network_client.virtual_wans.begin_create_or_update(resource_group_name=self.resource_group,
                                                                               virtual_wan_name=self.name,
                                                                               wan_parameters=self.body)
            if isinstance(response, LROPoller):
                response = self.get_poller_result(response)
        except Exception as exc:
            self.log('Error attempting to create the VirtualWan instance.')
            self.fail('Error creating the VirtualWan instance: {0}'.format(str(exc)))
        return response.as_dict()

    def delete_resource(self):
        try:
            response = self.network_client.virtual_wans.begin_delete(resource_group_name=self.resource_group,
                                                                     virtual_wan_name=self.name)
            if isinstance(response, LROPoller):
                self.get_poller_result(response)
        except Exception as e:
            self.log('Error attempting to delete the VirtualWan instance.')
            self.fail('Error deleting the VirtualWan instance: {0}'.format(str(e)))

        return True

    def get_resource(self):
        try:
            response = self.network_client.virtual_wans.get(resource_group_name=self.resource_group,
                                                            virtual_wan_name=self.name)
        except ResourceNotFoundError as e:
            return False
        return response.as_dict()


def main():
    AzureRMVirtualWan()


if __name__ == '__main__':
    main()



#!/usr/bin/python
#
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import absolute_import, division, print_function
__metaclass__ = type

DOCUMENTATION = '''
---
module: azure_rm_ml_model
version_added: "3.17.0"
short_description: Create, Archive or Restore an Azure Machine Learning Model
description:
    - Create, Archive or Restore an Azure Machine Learning Model.
options:
    name:
        description:
            - Name of the Azure ML model.
        type: str
        required: true
    resource_group:
        description:
            - Name of resource group.
        required: false
        type: str
    ml_workspace:
        description:
            - Name of the Azure ML workspace.
        required: false
        type: str
    ml_registry:
        description:
            - Name of the Azure ML registry.
        required: false
        type: str
    resource_definition:
        description:
            - Model definition in YAML
        type: str
        required: false
    version:
        description:
            - Data Asset version.
        type: str
    path:
        description:
            - Path to the model file(s). This can be either a local
              or a remote location. If specified, name and version
              must also be provided.
        type: str
    stage:
        description:
            - Stage of the model.
        type: str
        choices:
            - Staging
            - Production
            - Archived
            - Logged
            - Development
            - Deprecated
            - Retired
    datastore:
        description:
            - The datastore to upload the local artifact to.
        type: str
    description:
        description:
            - Description of the model.
        type: str
    type:
        description:
            - Type of the model.
        default: custom_model
        type: str
        choices:
            - custom_model
            - mlflow_model
            - triton_model
    state:
        description:
            - State of the Model. Use C(present) to create
              or update. C(archive) to archive, C(restore) to
              restore.
        default: present
        type: str
        choices:
            - present
            - archive
            - restore

extends_documentation_fragment:
    - azure.azcollection.azure
    - azure.azcollection.azure_tags

author:
    - Bill Peck (@p3ck)
'''

EXAMPLES = '''
- name: Create ML Model
  azure.azcollection.azure_rm_ml_model:
    name: modelxxxxxxxxxx
    resource_group: xxxxx-ml-model
    ml_workspace: workspace-xxxxxxxxxx
    version: 2
    stage: Production
    resource_definition: |
      $schema: https://azuremlschemas.azureedge.net/latest/model.schema.json
      path: ./files/mlflow-model/model.pkl
      description: Model created from local file.
'''

RETURN = '''
changed:
    description:
        - Whether the resource is changed.
    returned: always
    type: bool
    sample: false
ml_model:
    description:
        - Model that was just created or updated.
    returned: always
    type: dict
    sample: {
        "creation_context": {
            "created_at": "2026-03-25T17:48:05.464195+00:00",
            "created_by": "xxxxxxxx-xxxx-xxxx-xxxx-xxxxxxxxxxxx",
            "created_by_type": "Application",
            "last_modified_at": "2026-03-25T17:48:12.822846+00:00",
            "last_modified_by": "xxxxxxxx-xxxx-xxxx-xxxx-xxxxxxxxxxxx",
            "last_modified_by_type": "Application"
        },
        "description": "Model created from local file.",
        "id": "azureml:/subscriptions/xxxxxxxx-xxxx-xxxx-xxxx-xxxxxxxxxxxx/resourceGroups/xxxxx-ml-model/providers/Microsoft.MachineLearningServices/workspaces/workspace-xxxxxxxxxx/models/modelxxxxxxxxxx/versions/2",
        "name": "modelxxxxxxxxxx",
        "path": "azureml://subscriptions/xxxxxxxx-xxxx-xxxx-xxxx-xxxxxxxxxxxx/resourceGroups/xxxxx-ml-model/workspaces/workspace-xxxxxxxxxx/datastores/workspaceblobstore/paths/LocalUpload/xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx/model.pkl",
        "properties": {},
        "stage": "Production",
        "tags": {},
        "type": "custom_model",
        "version": "2"
    }
'''  # NOQA


try:
    import io
    import time
    from ansible_collections.azure.azcollection.plugins.module_utils.azure_rm_common_ml import MLClientCommon
    from azure.core.exceptions import ResourceNotFoundError
    from azure.ai.ml.entities._load_functions import load_model
    from azure.ai.ml.entities._assets import Model
except ImportError:
    # This is handled in azure_rm_common
    pass


class AzureRMMLModel(MLClientCommon):

    UPDATE_OVERRIDES = ['stage']

    PARAM_OVERRIDES = UPDATE_OVERRIDES + ['name',
                                          'version',
                                          'path',
                                          'type',
                                          'description',
                                          'tags',
                                          'datastore']

    def __init__(self):

        self.module_arg_spec = dict(
            name=dict(
                type='str',
                required=True,
            ),
            resource_definition=dict(
                type='str',
                required=False,
            ),
            resource_group=dict(
                type='str',
                required=False,
            ),
            ml_workspace=dict(
                type='str',
                required=False,
            ),
            ml_registry=dict(
                type='str',
                required=False,
            ),
            version=dict(
                type='str',
                required=False,
            ),
            path=dict(
                type='str',
                required=False,
            ),
            stage=dict(
                type='str',
                required=False,
                choices=['Staging', 'Production', 'Archived',
                         'Logged', 'Development', 'Deprecated', 'Retired'],
            ),
            datastore=dict(
                type='str',
                required=False,
            ),
            description=dict(
                type='str',
                required=False,
            ),
            type=dict(
                type='str',
                default='custom_model',
                choices=['custom_model', 'mlflow_model', 'triton_model'],
            ),
            state=dict(
                type='str',
                default='present',
                choices=['present', 'archive', 'restore'],
            ),
        )

        self._client = None
        self.ml_registry = None
        self.state = None

        self.results = dict(
            changed=False,
            compare=[],
            ml_model={}
        )

        required_one_of = [('ml_workspace', 'ml_registry')]
        mutually_exclusive = [('ml_workspace', 'ml_registry')]
        required_together = [('ml_workspace', 'resource_group')]

        super(AzureRMMLModel, self).__init__(self.module_arg_spec,
                                             supports_check_mode=True,
                                             required_one_of=required_one_of,
                                             mutually_exclusive=mutually_exclusive,
                                             required_together=required_together,
                                             )

    def exec_module(self, **kwargs):

        for key in list(self.module_arg_spec.keys()) + ['tags']:
            setattr(self, key, kwargs[key])

        changed = False

        if self.state == 'present':
            ml_model_info = self.get(name=self.name,
                                     version=self.version)

            if ml_model_info:
                # Update
                params_override = self.update_params(kwargs,
                                                     self.UPDATE_OVERRIDES,
                                                     ml_model_info)

                # Generate ml_model based on ansible args passed in.
                ml_model = Model._load(data=[],
                                       params_override=params_override)

                ml_model = load_model(None, params_override=params_override)
                ml_model_info = self.entity_to_dict(ml_model_info)
                changed = not self.default_compare({},
                                                   self.entity_to_dict(ml_model),
                                                   ml_model_info,
                                                   '',
                                                   self.results)
                if changed and not self.check_mode:
                    response = self.client.create_or_update(ml_model)
                    ml_model_info = self.entity_to_dict(response)
            else:
                # Create
                params_override = self.update_params(kwargs, self.PARAM_OVERRIDES)

                # Generate ml_model based on ansible args passed in.
                if self.resource_definition:
                    resource_definition = io.StringIO(self.resource_definition)
                    ml_model = load_model(source=resource_definition,
                                          params_override=params_override)
                else:
                    ml_model = Model(
                        name=self.name,
                        version=self.version,
                        path=self.path,
                        description=self.description,
                        tags=self.tags,
                        stage=self.stage,
                        type=self.type,
                        datastore=self.datastore,
                    )

                changed = True
                if not self.check_mode:
                    response = self.client.create_or_update(ml_model)
                    ml_model_info = self.entity_to_dict(response)
        elif self.state == 'archive':
            # Archive
            list_view_type = self.get_list_view_type('active')
            results = self.client.models.list(list_view_type=list_view_type)
            ml_models = [x.name for x in results]

            if self.name in ml_models:
                changed = True
                if not self.check_mode:
                    self.archive(name=self.name)
            ml_model_info = None
        elif self.state == 'restore':
            # Restore
            list_view_type = self.get_list_view_type('archived')
            results = self.client.models.list(list_view_type=list_view_type)
            ml_models = [x.name for x in results]

            if self.name in ml_models:
                changed = True
                if not self.check_mode:
                    self.restore(name=self.name)
                    # If we query too quick it hasn't fully restored.
                    # Seems the api shouldn't return until it's finished,
                    # but here we are.
                    time.sleep(1)
            ml_model_info = self.get(name=self.name,
                                     label='latest',
                                     as_dict=True)

        self.results['ml_model'] = ml_model_info
        self.results['changed'] = changed
        return self.results

    def update_params(self, kwargs, param_overrides, ml_model=None):
        params_override = []

        # If ml_model is defined use those values as a base.
        # Update can't change some of these values but we need
        # to keep the same values as before the update.
        if ml_model:
            items = self.entity_to_dict(ml_model).items()
            params_override = [{k: v} for k, v in items]

        # Update values based on what param_overrides says we can update.
        for key in list(self.module_arg_spec.keys()):
            if kwargs[key] is not None and key in param_overrides:
                params_override.append({key: kwargs[key]})
        return params_override

    def get(self, name, version=None, label=None, as_dict=False):
        try:
            # Attempt to retrieve ml_model based on name
            ml_model_info = self.client.models.get(name=name,
                                                   version=version,
                                                   label=label)
            if as_dict:
                ml_model_info = self.entity_to_dict(ml_model_info)
        except ResourceNotFoundError:
            ml_model_info = None
        return ml_model_info

    def archive(self, name=None):
        try:
            self.client.models.archive(name=name)
        except Exception as e:  # pylint: disable=broad-exception-caught
            self.fail("Error archive the model instance: {0}".format(str(e)))

    def restore(self, name=None):
        try:
            self.client.models.restore(name=name)
        except Exception as e:  # pylint: disable=broad-exception-caught
            self.fail("Error restore the model instance: {0}".format(str(e)))


def main():
    AzureRMMLModel()


if __name__ == '__main__':
    main()



# Derived from ansible/plugins/connection/proxmox_pct_remote.py (c) 2024 Nils Stein (@mietzen) <github.nstein@mailbox.org>
# Derived from ansible/plugins/connection/paramiko_ssh.py (c) 2012, Michael DeHaan <michael.dehaan@gmail.com>
# Copyright (c) 2025 Rui Lopes (@rgl) <ruilopes.com>
# Copyright (c) 2025 Ansible Project
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later

from __future__ import annotations

DOCUMENTATION = r"""
author: Rui Lopes (@rgl) <ruilopes.com>
name: wsl
short_description: Run tasks in WSL distribution using wsl.exe CLI using SSH
requirements:
  - paramiko
description:
  - Run commands or put/fetch files to an existing WSL distribution using wsl.exe CLI using SSH.
  - Uses the Python SSH implementation (Paramiko) to connect to the WSL host.
version_added: "10.6.0"
options:
  remote_addr:
    description:
      - Address of the remote target.
    default: inventory_hostname
    type: string
    vars:
      - name: inventory_hostname
      - name: ansible_host
      - name: ansible_ssh_host
      - name: ansible_paramiko_host
  port:
    description: Remote port to connect to.
    type: int
    default: 22
    ini:
      - section: defaults
        key: remote_port
      - section: paramiko_connection
        key: remote_port
    env:
      - name: ANSIBLE_REMOTE_PORT
      - name: ANSIBLE_REMOTE_PARAMIKO_PORT
    vars:
      - name: ansible_port
      - name: ansible_ssh_port
      - name: ansible_paramiko_port
    keyword:
      - name: port
  remote_user:
    description:
      - User to login/authenticate as.
      - Can be set from the CLI with the C(--user) or C(-u) options.
    type: string
    vars:
      - name: ansible_user
      - name: ansible_ssh_user
      - name: ansible_paramiko_user
    env:
      - name: ANSIBLE_REMOTE_USER
      - name: ANSIBLE_PARAMIKO_REMOTE_USER
    ini:
      - section: defaults
        key: remote_user
      - section: paramiko_connection
        key: remote_user
    keyword:
      - name: remote_user
  password:
    description:
      - Secret used to either login the SSH server or as a passphrase for SSH keys that require it.
      - Can be set from the CLI with the C(--ask-pass) option.
    type: string
    vars:
      - name: ansible_password
      - name: ansible_ssh_pass
      - name: ansible_ssh_password
      - name: ansible_paramiko_pass
      - name: ansible_paramiko_password
  use_rsa_sha2_algorithms:
    description:
      - Whether or not to enable RSA SHA2 algorithms for pubkeys and hostkeys.
      - On paramiko versions older than 2.9, this only affects hostkeys.
      - For behavior matching paramiko<2.9 set this to V(false).
    vars:
      - name: ansible_paramiko_use_rsa_sha2_algorithms
    ini:
      - {key: use_rsa_sha2_algorithms, section: paramiko_connection}
    env:
      - {name: ANSIBLE_PARAMIKO_USE_RSA_SHA2_ALGORITHMS}
    default: true
    type: boolean
  host_key_auto_add:
    description: "Automatically add host keys to C(~/.ssh/known_hosts)."
    env:
      - name: ANSIBLE_PARAMIKO_HOST_KEY_AUTO_ADD
    ini:
      - key: host_key_auto_add
        section: paramiko_connection
    type: boolean
  look_for_keys:
    default: true
    description: "Set to V(false) to disable searching for private key files in C(~/.ssh/)."
    env:
      - name: ANSIBLE_PARAMIKO_LOOK_FOR_KEYS
    ini:
      - {key: look_for_keys, section: paramiko_connection}
    type: boolean
  proxy_command:
    default: ""
    description:
      - Proxy information for running the connection through a jumphost.
      - This option is supported by paramiko version 1.9.0 or newer.
    type: string
    env:
      - name: ANSIBLE_PARAMIKO_PROXY_COMMAND
    ini:
      - {key: proxy_command, section: paramiko_connection}
    vars:
      - name: ansible_paramiko_proxy_command
  record_host_keys:
    default: true
    description: "Save the host keys to a file."
    env:
      - name: ANSIBLE_PARAMIKO_RECORD_HOST_KEYS
    ini:
      - section: paramiko_connection
        key: record_host_keys
    type: boolean
  host_key_checking:
    description: "Set this to V(false) if you want to avoid host key checking by the underlying tools Ansible uses to connect
      to the host."
    type: boolean
    default: true
    env:
      - name: ANSIBLE_HOST_KEY_CHECKING
      - name: ANSIBLE_SSH_HOST_KEY_CHECKING
      - name: ANSIBLE_PARAMIKO_HOST_KEY_CHECKING
    ini:
      - section: defaults
        key: host_key_checking
      - section: paramiko_connection
        key: host_key_checking
    vars:
      - name: ansible_host_key_checking
      - name: ansible_ssh_host_key_checking
      - name: ansible_paramiko_host_key_checking
  use_persistent_connections:
    description: "Toggles the use of persistence for connections."
    type: boolean
    default: false
    env:
      - name: ANSIBLE_USE_PERSISTENT_CONNECTIONS
    ini:
      - section: defaults
        key: use_persistent_connections
  banner_timeout:
    type: float
    default: 30
    description:
      - Configures, in seconds, the amount of time to wait for the SSH banner to be presented.
      - This option is supported by paramiko version 1.15.0 or newer.
    ini:
      - section: paramiko_connection
        key: banner_timeout
    env:
      - name: ANSIBLE_PARAMIKO_BANNER_TIMEOUT
  timeout:
    type: int
    default: 10
    description:
      - Number of seconds until the plugin gives up on failing to establish a TCP connection.
      - This option is supported by paramiko version 2.2.0 or newer.
    ini:
      - section: defaults
        key: timeout
      - section: ssh_connection
        key: timeout
      - section: paramiko_connection
        key: timeout
    env:
      - name: ANSIBLE_TIMEOUT
      - name: ANSIBLE_SSH_TIMEOUT
      - name: ANSIBLE_PARAMIKO_TIMEOUT
    vars:
      - name: ansible_ssh_timeout
      - name: ansible_paramiko_timeout
    cli:
      - name: timeout
  lock_file_timeout:
    type: int
    default: 60
    description: Number of seconds until the plugin gives up on trying to write a lock file when writing SSH known host keys.
    vars:
      - name: ansible_lock_file_timeout
    env:
      - name: ANSIBLE_LOCK_FILE_TIMEOUT
  private_key_file:
    description:
      - Path to private key file to use for authentication.
    type: path
    ini:
      - section: defaults
        key: private_key_file
      - section: paramiko_connection
        key: private_key_file
    env:
      - name: ANSIBLE_PRIVATE_KEY_FILE
      - name: ANSIBLE_PARAMIKO_PRIVATE_KEY_FILE
    vars:
      - name: ansible_private_key_file
      - name: ansible_ssh_private_key_file
      - name: ansible_paramiko_private_key_file
    cli:
      - name: private_key_file
        option: "--private-key"
  user_known_hosts_file:
    description:
      - Path to the user known hosts file.
      - Used to verify the ssh hosts keys.
    type: path
    default: ~/.ssh/known_hosts
    ini:
      - section: paramiko_connection
        key: user_known_hosts_file
    vars:
      - name: ansible_paramiko_user_known_hosts_file
  wsl_distribution:
    description:
      - WSL distribution name.
    type: string
    required: true
    vars:
      - name: wsl_distribution
  wsl_remote_ssh_shell_type:
    description:
      - The shell type expected in the SSH session (not inside the WSL session).
      - See also C(ansible_shell_type).
    type: string
    choices:
      - cmd
      - powershell
    default: cmd
    vars:
      - name: wsl_remote_ssh_shell_type
    version_added: 12.2.0
  wsl_user:
    description:
      - WSL distribution user.
    type: string
    vars:
      - name: wsl_user
  become_user:
    description:
      - WSL distribution user.
    type: string
    default: root
    vars:
      - name: become_user
      - name: ansible_become_user
  become:
    description:
      - Whether to use the user defined by O(become_user).
    type: bool
    default: false
    vars:
      - name: become
      - name: ansible_become
"""

EXAMPLES = r"""
# ------------------------
# Inventory: inventory.yml
# ------------------------
---
all:
  children:
    wsl:
      hosts:
        example-wsl-ubuntu:
          ansible_host: 10.0.0.10
          wsl_distribution: ubuntu
          wsl_user: ubuntu
      vars:
        ansible_connection: community.general.wsl
        ansible_user: vagrant
# ----------------------
# Playbook: playbook.yml
# ----------------------
---
- name: WSL Example
  hosts: wsl
  gather_facts: true
  become: true
  tasks:
    - name: Ping
      ansible.builtin.ping:
    - name: Id (with become false)
      become: false
      changed_when: false
      args:
        executable: /bin/bash
      ansible.builtin.shell: |
        exec 2>&1
        set -x
        echo "$0"
        pwd
        id
    - name: Id (with become true)
      changed_when: false
      args:
        executable: /bin/bash
      ansible.builtin.shell: |
        exec 2>&1
        set -x
        echo "$0"
        pwd
        id
    - name: Reboot
      ansible.builtin.reboot:
        boot_time_command: systemctl show -p ActiveEnterTimestamp init.scope
"""

import io
import os
import pathlib
import shlex
import tempfile
import traceback
import typing as t
from binascii import hexlify
from subprocess import list2cmdline

from ansible.errors import (
    AnsibleAuthenticationFailure,
    AnsibleConnectionFailure,
    AnsibleError,
)
from ansible.module_utils.common.text.converters import to_bytes, to_native, to_text
from ansible.playbook.play_context import PlayContext
from ansible.plugins.connection import ConnectionBase
from ansible.utils.display import Display
from ansible.utils.path import makedirs_safe

from ansible_collections.community.general.plugins.module_utils._filelock import FileLock, LockTimeout
from ansible_collections.community.general.plugins.module_utils._version import LooseVersion

PARAMIKO_IMPORT_ERR: str | None
try:
    import paramiko
    from paramiko import MissingHostKeyPolicy

    PARAMIKO_IMPORT_ERR = None
except ImportError:
    PARAMIKO_IMPORT_ERR = traceback.format_exc()
    MissingHostKeyPolicy = object  # type: ignore


display = Display()


def authenticity_msg(hostname: str, ktype: str, fingerprint: bytes) -> str:
    msg = f"""
    paramiko: The authenticity of host '{hostname}' can't be established.
    The {ktype} key fingerprint is {to_text(fingerprint)}.
    Are you sure you want to continue connecting (yes/no)?
    """
    return msg


class MyAddPolicy(MissingHostKeyPolicy):
    """
    Based on AutoAddPolicy in paramiko so we can determine when keys are added

    and also prompt for input.

    Policy for automatically adding the hostname and new host key to the
    local L{HostKeys} object, and saving it. This is used by L{SSHClient}.
    """

    def __init__(self, connection: Connection) -> None:
        self.connection = connection
        self._options = connection._options

    def missing_host_key(self, client: paramiko.SSHClient, hostname: str, key: paramiko.PKey) -> None:
        if all((self.connection.get_option("host_key_checking"), not self.connection.get_option("host_key_auto_add"))):
            fingerprint = hexlify(key.get_fingerprint())
            ktype = key.get_name()

            if self.connection.get_option("use_persistent_connections") or self.connection.force_persistence:
                # don't print the prompt string since the user cannot respond
                # to the question anyway
                raise AnsibleError(authenticity_msg(hostname, ktype, fingerprint)[1:92])

            inp = to_text(
                display.prompt_until(authenticity_msg(hostname, ktype, fingerprint), private=False),
                errors="surrogate_or_strict",
            )

            if inp.lower() not in ["yes", "y", ""]:
                raise AnsibleError("host connection rejected by user")

        key._added_by_ansible_this_time = True  # type: ignore

        # existing implementation below:
        client._host_keys.add(hostname, key.get_name(), key)  # type: ignore[attr-defined]  # TODO: figure out what _host_keys is!

        # host keys are actually saved in close() function below
        # in order to control ordering.


class Connection(ConnectionBase):
    """SSH based connections (paramiko) to WSL"""

    transport = "community.general.wsl"
    _log_channel: str | None = None

    def __init__(
        self, play_context: PlayContext, new_stdin: io.TextIOWrapper | None = None, *args: t.Any, **kwargs: t.Any
    ):
        super().__init__(play_context, new_stdin, *args, **kwargs)

    def _set_log_channel(self, name: str) -> None:
        """Mimic paramiko.SSHClient.set_log_channel"""
        self._log_channel = name

    def _parse_proxy_command(self, port: int = 22) -> dict[str, t.Any]:
        proxy_command = self.get_option("proxy_command") or None

        sock_kwarg = {}
        if proxy_command:
            replacers: dict[str, str] = {
                "%h": self.get_option("remote_addr"),
                "%p": str(port),
                "%r": self.get_option("remote_user"),
            }
            for find, replace in replacers.items():
                proxy_command = proxy_command.replace(find, replace)
            try:
                sock_kwarg = {"sock": paramiko.ProxyCommand(proxy_command)}
                display.vvv(
                    f"CONFIGURE PROXY COMMAND FOR CONNECTION: {proxy_command}", host=self.get_option("remote_addr")
                )
            except AttributeError:
                display.warning(
                    "Paramiko ProxyCommand support unavailable. "
                    "Please upgrade to Paramiko 1.9.0 or newer. "
                    "Not using configured ProxyCommand"
                )

        return sock_kwarg

    def _connect(self) -> Connection:
        """activates the connection object"""

        if PARAMIKO_IMPORT_ERR is not None:
            raise AnsibleError(f"paramiko is not installed: {to_native(PARAMIKO_IMPORT_ERR)}")

        port = self.get_option("port")
        display.vvv(
            f"ESTABLISH PARAMIKO SSH CONNECTION FOR USER: {self.get_option('remote_user')} on PORT {to_text(port)} TO {self.get_option('remote_addr')}",
            host=self.get_option("remote_addr"),
        )

        ssh = paramiko.SSHClient()

        # Set pubkey and hostkey algorithms to disable, the only manipulation allowed currently
        # is keeping or omitting rsa-sha2 algorithms
        # default_keys: t.Tuple[str] = ()
        paramiko_preferred_pubkeys = getattr(paramiko.Transport, "_preferred_pubkeys", ())
        paramiko_preferred_hostkeys = getattr(paramiko.Transport, "_preferred_keys", ())
        use_rsa_sha2_algorithms = self.get_option("use_rsa_sha2_algorithms")
        disabled_algorithms: dict[str, t.Iterable[str]] = {}
        if not use_rsa_sha2_algorithms:
            if paramiko_preferred_pubkeys:
                disabled_algorithms["pubkeys"] = tuple(a for a in paramiko_preferred_pubkeys if "rsa-sha2" in a)
            if paramiko_preferred_hostkeys:
                disabled_algorithms["keys"] = tuple(a for a in paramiko_preferred_hostkeys if "rsa-sha2" in a)

        # override paramiko's default logger name
        if self._log_channel is not None:
            ssh.set_log_channel(self._log_channel)

        self.keyfile = os.path.expanduser(self.get_option("user_known_hosts_file"))

        if self.get_option("host_key_checking"):
            for ssh_known_hosts in ("/etc/ssh/ssh_known_hosts", "/etc/openssh/ssh_known_hosts", self.keyfile):
                try:
                    ssh.load_system_host_keys(ssh_known_hosts)
                    break
                except OSError:
                    pass  # file was not found, but not required to function
                except paramiko.hostkeys.InvalidHostKey as e:
                    raise AnsibleConnectionFailure(f"Invalid host key: {to_text(e.line)}") from e
            try:
                ssh.load_system_host_keys()
            except paramiko.hostkeys.InvalidHostKey as e:
                raise AnsibleConnectionFailure(f"Invalid host key: {to_text(e.line)}") from e

        ssh_connect_kwargs = self._parse_proxy_command(port)
        ssh.set_missing_host_key_policy(MyAddPolicy(self))
        conn_password = self.get_option("password")
        allow_agent = True

        if conn_password is not None:
            allow_agent = False

        try:
            key_filename = None
            if self.get_option("private_key_file"):
                key_filename = os.path.expanduser(self.get_option("private_key_file"))

            # paramiko 2.2 introduced auth_timeout parameter
            if LooseVersion(paramiko.__version__) >= LooseVersion("2.2.0"):
                ssh_connect_kwargs["auth_timeout"] = self.get_option("timeout")

            # paramiko 1.15 introduced banner timeout parameter
            if LooseVersion(paramiko.__version__) >= LooseVersion("1.15.0"):
                ssh_connect_kwargs["banner_timeout"] = self.get_option("banner_timeout")

            ssh.connect(
                self.get_option("remote_addr").lower(),
                username=self.get_option("remote_user"),
                allow_agent=allow_agent,
                look_for_keys=self.get_option("look_for_keys"),
                key_filename=key_filename,
                password=conn_password,
                timeout=self.get_option("timeout"),
                port=port,
                disabled_algorithms=disabled_algorithms,
                **ssh_connect_kwargs,
            )
        except paramiko.ssh_exception.BadHostKeyException as e:
            raise AnsibleConnectionFailure(f"host key mismatch for {to_text(e.hostname)}") from e
        except paramiko.ssh_exception.AuthenticationException as e:
            msg = f"Failed to authenticate: {e}"
            raise AnsibleAuthenticationFailure(msg) from e
        except Exception as e:
            msg = to_text(e)
            if "PID check failed" in msg:
                raise AnsibleError(
                    "paramiko version issue, please upgrade paramiko on the machine running ansible"
                ) from e
            elif "Private key file is encrypted" in msg:
                msg = (
                    f"ssh {self.get_option('remote_user')}@{self.get_options('remote_addr')}:{port} : "
                    f"{msg}\nTo connect as a different user, use -u <username>."
                )
                raise AnsibleConnectionFailure(msg) from e
            else:
                raise AnsibleConnectionFailure(msg) from e
        self.ssh = ssh
        self._connected = True
        return self

    def _any_keys_added(self) -> bool:
        for host_keys in self.ssh._host_keys.values():  # type: ignore[attr-defined]  # TODO: figure out what _host_keys is!
            for key in host_keys.values():
                added_this_time = getattr(key, "_added_by_ansible_this_time", False)
                if added_this_time:
                    return True
        return False

    def _save_ssh_host_keys(self, filename: str) -> None:
        """
        not using the paramiko save_ssh_host_keys function as we want to add new SSH keys at the bottom so folks
        don't complain about it :)
        """

        if not self._any_keys_added():
            return

        path = os.path.expanduser("~/.ssh")
        makedirs_safe(path)

        with open(filename, "w") as f:
            for hostname, keys in self.ssh._host_keys.items():  # type: ignore[attr-defined]  # TODO: figure out what _host_keys is!
                for keytype, key in keys.items():
                    # was f.write
                    added_this_time = getattr(key, "_added_by_ansible_this_time", False)
                    if not added_this_time:
                        f.write(f"{hostname} {keytype} {key.get_base64()}\n")

            for hostname, keys in self.ssh._host_keys.items():  # type: ignore[attr-defined]  # TODO: figure out what _host_keys is!
                for keytype, key in keys.items():
                    added_this_time = getattr(key, "_added_by_ansible_this_time", False)
                    if added_this_time:
                        f.write(f"{hostname} {keytype} {key.get_base64()}\n")

    def _build_wsl_command(self, cmd: str) -> str:
        wsl_distribution = self.get_option("wsl_distribution")
        become = self.get_option("become")
        become_user = self.get_option("become_user")
        wsl_remote_ssh_shell_type = self.get_option("wsl_remote_ssh_shell_type")
        is_integration_test = os.getenv("_ANSIBLE_TEST_WSL_CONNECTION_PLUGIN_WAERI5TEPHEESHA2FAE8")
        if "%" in cmd:
            if wsl_remote_ssh_shell_type == "powershell":
                # there is no universal way to escape '%' here
                # if this is raised, add a workaround to allow the specific situation (if possible)
                raise AnsibleError("The command contains '%', cannot safely escape it for Powershell")
            else:
                cmd = cmd.replace("%", "^%")
        if become and become_user:
            wsl_user = become_user
        else:
            wsl_user = self.get_option("wsl_user")
        args = ["wsl.exe"]
        if wsl_remote_ssh_shell_type == "powershell" and not is_integration_test:
            # Powershell stop-parsing token, treat the rest as arguments to the native command wsl.exe
            args.append("--%")
        args.extend(["--distribution", wsl_distribution])
        if wsl_user:
            args.extend(["--user", wsl_user])
        args.extend(["--"])
        args.extend(shlex.split(cmd))
        if is_integration_test:
            return shlex.join(args)
        else:
            return list2cmdline(args)  # see https://github.com/python/cpython/blob/3.11/Lib/subprocess.py#L576

    def exec_command(self, cmd: str, in_data: bytes | None = None, sudoable: bool = True) -> tuple[int, bytes, bytes]:
        """run a command on inside a WSL distribution"""

        cmd = self._build_wsl_command(cmd)

        super().exec_command(cmd, in_data=in_data, sudoable=sudoable)  # type: ignore[safe-super]

        bufsize = 4096

        try:
            transport = self.ssh.get_transport()
            if transport is None:
                raise ValueError("Transport not available")
            transport.set_keepalive(5)
            chan = transport.open_session()
        except Exception as e:
            text_e = to_text(e)
            msg = "Failed to open session"
            if text_e:
                msg += f": {text_e}"
            raise AnsibleConnectionFailure(to_native(msg)) from e

        display.vvv(f"EXEC {cmd}", host=self.get_option("remote_addr"))

        cmd_b = to_bytes(cmd, errors="surrogate_or_strict")

        no_prompt_out = b""
        no_prompt_err = b""
        become_output = b""

        try:
            chan.exec_command(cmd_b)
            if self.become and self.become.expect_prompt():
                password_prompt = False
                become_success = False
                while not (become_success or password_prompt):
                    display.debug("Waiting for Privilege Escalation input")

                    chunk = chan.recv(bufsize)
                    display.debug(f"chunk is: {to_text(chunk)}")
                    if not chunk:
                        if b"unknown user" in become_output:
                            n_become_user = to_native(self.become.get_option("become_user"))
                            raise AnsibleError(f"user {n_become_user} does not exist")
                        else:
                            break
                            # raise AnsibleError('ssh connection closed waiting for password prompt')
                    become_output += chunk

                    # need to check every line because we might get lectured
                    # and we might get the middle of a line in a chunk
                    for line in become_output.splitlines(True):
                        if self.become.check_success(line):
                            become_success = True
                            break
                        elif self.become.check_password_prompt(line):
                            password_prompt = True
                            break

                if password_prompt:
                    if self.become:
                        become_pass = self.become.get_option("become_pass")
                        chan.sendall(to_bytes(f"{become_pass}\n", errors="surrogate_or_strict"))
                    else:
                        raise AnsibleError("A password is required but none was supplied")
                else:
                    no_prompt_out += become_output
                    no_prompt_err += become_output

            if in_data:
                for i in range(0, len(in_data), bufsize):
                    chan.send(in_data[i : i + bufsize])
                chan.shutdown_write()
            elif in_data == b"":
                chan.shutdown_write()

        except TimeoutError as e:
            raise AnsibleError(f"ssh timed out waiting for privilege escalation.\n{to_text(become_output)}") from e

        stdout = b"".join(chan.makefile("rb", bufsize))
        stderr = b"".join(chan.makefile_stderr("rb", bufsize))
        returncode = chan.recv_exit_status()

        # NB the full english error message is:
        #     'wsl.exe' is not recognized as an internal or external command,
        #     operable program or batch file.
        if "'wsl.exe' is not recognized" in stderr.decode("utf-8"):
            raise AnsibleError(f"wsl.exe not found in path of host: {to_text(self.get_option('remote_addr'))}")

        return (returncode, no_prompt_out + stdout, no_prompt_out + stderr)

    def put_file(self, in_path: str, out_path: str) -> None:
        """transfer a file from local to remote"""

        display.vvv(f"PUT {in_path} TO {out_path}", host=self.get_option("remote_addr"))
        try:
            with open(in_path, "rb") as f:
                data = f.read()
                returncode, stdout, stderr = self.exec_command(
                    f"{self._shell.executable} -c {self._shell.quote(f'cat > {out_path}')}",
                    in_data=data,
                    sudoable=False,
                )
            if returncode != 0:
                if "cat: not found" in stderr.decode("utf-8"):
                    raise AnsibleError(
                        f"cat not found in path of WSL distribution: {to_text(self.get_option('wsl_distribution'))}"
                    )
                raise AnsibleError(f"{to_text(stdout)}\n{to_text(stderr)}")
        except Exception as e:
            raise AnsibleError(f"error occurred while putting file from {in_path} to {out_path}!\n{to_text(e)}") from e

    def fetch_file(self, in_path: str, out_path: str) -> None:
        """save a remote file to the specified path"""

        display.vvv(f"FETCH {in_path} TO {out_path}", host=self.get_option("remote_addr"))
        try:
            returncode, stdout, stderr = self.exec_command(
                f"{self._shell.executable} -c {self._shell.quote(f'cat {in_path}')}", sudoable=False
            )
            if returncode != 0:
                if "cat: not found" in stderr.decode("utf-8"):
                    raise AnsibleError(
                        f"cat not found in path of WSL distribution: {to_text(self.get_option('wsl_distribution'))}"
                    )
                raise AnsibleError(f"{to_text(stdout)}\n{to_text(stderr)}")
            with open(out_path, "wb") as f:
                f.write(stdout)
        except Exception as e:
            raise AnsibleError(f"error occurred while fetching file from {in_path} to {out_path}!\n{to_text(e)}") from e

    def reset(self) -> None:
        """reset the connection"""

        if not self._connected:
            return
        self.close()
        self._connect()

    def close(self) -> None:
        """terminate the connection"""

        if self.get_option("host_key_checking") and self.get_option("record_host_keys") and self._any_keys_added():
            # add any new SSH host keys -- warning -- this could be slow
            # (This doesn't acquire the connection lock because it needs
            # to exclude only other known_hosts writers, not connections
            # that are starting up.)
            lockfile = os.path.basename(self.keyfile)
            dirname = os.path.dirname(self.keyfile)
            makedirs_safe(dirname)
            tmp_keyfile_name = None
            try:
                with FileLock().lock_file(lockfile, dirname, self.get_option("lock_file_timeout")):
                    # just in case any were added recently

                    self.ssh.load_system_host_keys()
                    self.ssh._host_keys.update(self.ssh._system_host_keys)  # type: ignore[attr-defined]  # TODO this is a HACK!

                    # gather information about the current key file, so
                    # we can ensure the new file has the correct mode/owner

                    key_dir = os.path.dirname(self.keyfile)
                    if os.path.exists(self.keyfile):
                        key_stat = os.stat(self.keyfile)
                        mode = key_stat.st_mode & 0o777
                        uid = key_stat.st_uid
                        gid = key_stat.st_gid
                    else:
                        mode = 0o644
                        uid = os.getuid()
                        gid = os.getgid()

                    # Save the new keys to a temporary file and move it into place
                    # rather than rewriting the file. We set delete=False because
                    # the file will be moved into place rather than cleaned up.

                    with tempfile.NamedTemporaryFile(dir=key_dir, delete=False) as tmp_keyfile:
                        tmp_keyfile_name = tmp_keyfile.name
                        os.chmod(tmp_keyfile_name, mode)
                        os.chown(tmp_keyfile_name, uid, gid)
                        self._save_ssh_host_keys(tmp_keyfile_name)

                    os.rename(tmp_keyfile_name, self.keyfile)
            except LockTimeout as e:
                raise AnsibleError(
                    f"writing lock file for {self.keyfile} ran in to the timeout of {self.get_option('lock_file_timeout')}s"
                ) from e
            except paramiko.hostkeys.InvalidHostKey as e:
                raise AnsibleConnectionFailure(f"Invalid host key: {e.line}") from e
            except Exception as e:
                # unable to save keys, including scenario when key was invalid
                # and caught earlier
                raise AnsibleError(f"error occurred while writing SSH host keys!\n{to_text(e)}") from e
            finally:
                if tmp_keyfile_name is not None:
                    pathlib.Path(tmp_keyfile_name).unlink(missing_ok=True)

        self.ssh.close()
        self._connected = False
